# Copyright (c) 2015-2019, Activision Publishing, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Assertion library for python unit testing with a fluent API"""

from __future__ import print_function
import os
import contextlib
import inspect
import logging
import sys
import types
from .base import BaseMixin
from .collection import CollectionMixin
from .contains import ContainsMixin
from .date import DateMixin
from .dict import DictMixin
from .dynamic import DynamicMixin
from .extracting import ExtractingMixin
from .exception import ExceptionMixin
from .file import FileMixin
from .helpers import HelpersMixin
from .numeric import NumericMixin
from .snapshot import SnapshotMixin
from .string import StringMixin

__version__ = '1.1'

__tracebackhide__ = True  # clean tracebacks via py.test integration
contextlib.__tracebackhide__ = True  # monkey patch contextlib with clean py.test tracebacks


# soft assertions
_soft_ctx = 0
_soft_err = []


@contextlib.contextmanager
def soft_assertions():
    """Create a soft assertion context.

    Normally, any assertion failure will halt test execution immediately by raising an error.
    Soft assertions are way to collect assertion failures (and failure messages) together, to be
    raised all at once at the end, without halting your test.

    Examples:
        Create a soft assertion context, and some failing tests::

            from assertpy import assert_that, soft_assertions

            with soft_assertions():
                assert_that('foo').is_length(4)
                assert_that('foo').is_empty()
                assert_that('foo').is_false()
                assert_that('foo').is_digit()
                assert_that('123').is_alpha()

        When the context ends, any assertion failures are collected together and a single
        ``AssertionError`` is raised::

            AssertionError: soft assertion failures:
            1. Expected <foo> to be of length <4>, but was <3>.
            2. Expected <foo> to be empty string, but was not.
            3. Expected <False>, but was not.
            4. Expected <foo> to contain only digits, but did not.
            5. Expected <123> to contain only alphabetic chars, but did not.

    Note:
        The soft assertion context only collects *assertion* failures, other errors such as
        ``TypeError`` or ``ValueError`` are always raised immediately.  Triggering an explicit test
        failure with :meth:`fail` will similarly halt execution immediately.  If you need more
        forgiving behavior, use :meth:`soft_fail` to add a failure message without halting test
        execution.
    """
    global _soft_ctx
    global _soft_err

    # init ctx
    if _soft_ctx == 0:
        _soft_err = []
    _soft_ctx += 1

    try:
        yield
    finally:
        # reset ctx
        _soft_ctx -= 1

    if _soft_err and _soft_ctx == 0:
        out = 'soft assertion failures:'
        for i, msg in enumerate(_soft_err):
            out += '\n%d. %s' % (i+1, msg)
        # reset msg, then raise
        _soft_err = []
        raise AssertionError(out)


# factory methods
def assert_that(val, description=''):
    """Set the value to be tested, plus an optional description, and allow assertions to be called.

    This is a factory method for the :class:`AssertionBuilder`, and the single most important
    method in all of assertpy.

    Args:
        val: the value to be tested (aka the actual value)
        description (str, optional): the extra error message description. Defaults to ``''``
            (aka empty string)

    Examples:
        Just import it once at the top of your test file, and away you go...::

            from assertpy import assert_that

            def test_something():
                assert_that(1 + 2).is_equal_to(3)
                assert_that('foobar').is_length(6).starts_with('foo').ends_with('bar')
                assert_that(['a', 'b', 'c']).contains('a').does_not_contain('x')
    """
    global _soft_ctx
    if _soft_ctx:
        return _builder(val, description, 'soft')
    return _builder(val, description)


def assert_warn(val, description='', logger=None):
    """Set the value to be tested, and optional description and logger, and allow assertions to be
    called, but never fail, only log warnings.

    This is a factory method for the :class:`AssertionBuilder`, but unlike :meth:`assert_that` an
    `AssertionError` is never raised, and execution is never halted.  Instead, any assertion failures
    results in a warning message being logged. Uses the given logger, or defaults to a simple logger
    that prints warnings to ``stdout``.


    Args:
        val: the value to be tested (aka the actual value)
        description (str, optional): the extra error message description. Defaults to ``''``
            (aka empty string)
        logger (Logger, optional): the logger for warning message on assertion failure. Defaults to ``None``
            (aka use the default simple logger that prints warnings to ``stdout``)

    Examples:
        Usage::

            from assertpy import assert_warn

            assert_warn('foo').is_length(4)
            assert_warn('foo').is_empty()
            assert_warn('foo').is_false()
            assert_warn('foo').is_digit()
            assert_warn('123').is_alpha()

        Even though all of the above assertions fail, ``AssertionError`` is never raised and
        test execution is never halted.  Instead, the failed assertions merely log the following
        warning messages to ``stdout``::

            2019-10-27 20:00:35 WARNING [test_foo.py:23]: Expected <foo> to be of length <4>, but was <3>.
            2019-10-27 20:00:35 WARNING [test_foo.py:24]: Expected <foo> to be empty string, but was not.
            2019-10-27 20:00:35 WARNING [test_foo.py:25]: Expected <False>, but was not.
            2019-10-27 20:00:35 WARNING [test_foo.py:26]: Expected <foo> to contain only digits, but did not.
            2019-10-27 20:00:35 WARNING [test_foo.py:27]: Expected <123> to contain only alphabetic chars, but did not.

    Tip:
        Use :meth:`assert_warn` if and only if you have a *really* good reason to log assertion
        failures instead of failing.
    """
    return _builder(val, description, 'warn', logger=logger)


def fail(msg=''):
    """Force immediate test failure with the given message.

    Args:
        msg (str, optional): the failure message.  Defaults to ``''``

    Examples:
        Fail a test::

            from assertpy import assert_that, fail

            def test_fail():
                fail('forced fail!')

        If you wanted to test for a known failure, here is a useful pattern::

            import operator

            def test_adder_bad_arg():
                try:
                    operator.add(1, 'bad arg')
                    fail('should have raised error')
                except TypeError as e:
                    assert_that(str(e)).contains('unsupported operand')
    """
    raise AssertionError('Fail: %s!' % msg if msg else 'Fail!')


def soft_fail(msg=''):
    """Within a :meth:`soft_assertions` context, append the failure message to the soft error list,
    but do not halt test execution.

    Otherwise, outside the context, acts identical to :meth:`fail` and forces immediate test
    failure with the given message.

    Args:
        msg (str, optional): the failure message.  Defaults to ``''``

    Examples:
        Failing soft assertions::

            from assertpy import assert_that, soft_assertions, soft_fail

            with soft_assertions():
                assert_that(1).is_equal_to(2)
                soft_fail('my message')
                assert_that('foo').is_equal_to('bar')

        Fails, and outputs the following soft error list::

            AssertionError: soft assertion failures:
            1. Expected <1> to be equal to <2>, but was not.
            2. Fail: my message!
            3. Expected <foo> to be equal to <bar>, but was not.

    """
    global _soft_ctx
    if _soft_ctx:
        global _soft_err
        _soft_err.append('Fail: %s!' % msg if msg else 'Fail!')
        return
    fail(msg)


# assertion extensions
_extensions = {}


def add_extension(func):
    """Add a new user-defined custom assertion to assertpy.

    Once the assertion is registered with assertpy, use it like any other assertion.  Pass val to
    :meth:`assert_that`, and then call it.

    Args:
        func: the assertion function (to be added)

    Examples:
        Usage::

            from assertpy import add_extension

            def is_5(self):
                if self.val != 5:
                    self.error(f'{self.val} is NOT 5!')
                return self

            add_extension(is_5)

            def test_5():
                assert_that(5).is_5()

            def test_6():
                assert_that(6).is_5()  # fails
                # 6 is NOT 5!
    """
    if not callable(func):
        raise TypeError('func must be callable')
    _extensions[func.__name__] = func


def remove_extension(func):
    """Remove a user-defined custom assertion.

    Args:
        func: the assertion function (to be removed)

    Examples:
        Usage::

            from assertpy import remove_extension

            remove_extension(is_5)
    """
    if not callable(func):
        raise TypeError('func must be callable')
    if func.__name__ in _extensions:
        del _extensions[func.__name__]


def _builder(val, description='', kind=None, expected=None, logger=None):
    """Internal helper to build a new :class:`AssertionBuilder` instance and glue on any extension methods."""
    ab = AssertionBuilder(val, description, kind, expected, logger)
    if _extensions:
        # glue extension method onto new builder instance
        for name, func in _extensions.items():
            meth = types.MethodType(func, ab)
            setattr(ab, name, meth)
    return ab


# warnings
class WarningLoggingAdapter(logging.LoggerAdapter):
    """Logging adapter to unwind the stack to get the correct callee filename and line number."""

    def process(self, msg, kwargs):
        def _unwind(frame, fn='assert_warn'):
            if frame and fn in frame.f_code.co_names:
                return frame
            return _unwind(frame.f_back, fn)

        frame = _unwind(inspect.currentframe())
        lineno = frame.f_lineno
        filename = os.path.basename(frame.f_code.co_filename)
        return '[%s:%d]: %s' % (filename, lineno, msg), kwargs


_logger = logging.getLogger('assertpy')
_handler = logging.StreamHandler(sys.stdout)
_handler.setLevel(logging.WARNING)
_format = logging.Formatter('%(asctime)s %(levelname)s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
_handler.setFormatter(_format)
_logger.addHandler(_handler)
_default_logger = WarningLoggingAdapter(_logger, None)


class AssertionBuilder(
    StringMixin,
    SnapshotMixin,
    NumericMixin,
    HelpersMixin,
    FileMixin,
    ExtractingMixin,
    ExceptionMixin,
    DynamicMixin,
    DictMixin,
    DateMixin,
    ContainsMixin,
    CollectionMixin,
    BaseMixin,
    object
):
    """The main assertion class.  Never call the constructor directly, always use the
    :meth:`assert_that` helper instead.  Or if you just want warning messages, use the
    :meth:`assert_warn` helper.

    Args:
        val: the value to be tested (aka the actual value)
        description (str, optional): the extra error message description.  Defaults to ``''``
            (aka empty string)
        kind (str, optional): the kind of assertions, one of ``None``, ``soft``, or ``warn``.
            Defaults to ``None``
        expected (Error, optional): the expected exception.  Defaults to ``None``
        logger (Logger, optional): the logger for warning messages.  Defaults to ``None``
    """

    def __init__(self, val, description='', kind=None, expected=None, logger=None):
        """Never call this constructor directly."""
        self.val = val
        self.description = description
        self.kind = kind
        self.expected = expected
        self.logger = logger if logger else _default_logger

    def builder(self, val, description='', kind=None, expected=None, logger=None):
        """Helper to build a new :class:`AssertionBuilder` instance. Use this only if not chaining to ``self``.

        Args:
            val: the value to be tested (aka the actual value)
            description (str, optional): the extra error message description.  Defaults to ``''``
                (aka empty string)
            kind (str, optional): the kind of assertions, one of ``None``, ``soft``, or ``warn``.
                Defaults to ``None``
            expected (Error, optional): the expected exception.  Defaults to ``None``
            logger (Logger, optional): the logger for warning messages.  Defaults to ``None``
        """
        return _builder(val, description, kind, expected, logger)

    def error(self, msg):
        """Helper to raise an ``AssertionError`` with the given message.

        If an error description is set by :meth:`~assertpy.base.BaseMixin.described_as`, then that
        description is prepended to the error message.

        Args:
            msg: the error message

        Examples:
            Used to fail an assertion::

                if self.val != other:
                    self.error('Expected <%s> to be equal to <%s>, but was not.' % (self.val, other))

        Raises:
            AssertionError: always raised unless ``kind`` is ``warn`` (as set when using an
                :meth:`assert_warn` assertion) or ``kind`` is ``soft`` (as set when inside a
                :meth:`soft_assertions` context).
        """
        out = '%s%s' % ('[%s] ' % self.description if len(self.description) > 0 else '', msg)
        if self.kind == 'warn':
            self.logger.warning(out)
            return self
        elif self.kind == 'soft':
            global _soft_err
            _soft_err.append(out)
            return self
        else:
            raise AssertionError(out)



class RunningException(Exception):
    def __init__(self, message: str):
        super().__init__()
        self.message = message

    def __str__(self):
        return f'ERROR:\n' \
               f'Reason: {self.message}'


class UserException(Exception):
    def __init__(self, message: str):
        super().__init__()
        self.message = message

    def __str__(self):
        return f'ERROR:\n' \
               f'Reason: {self.message}'


class BackupException(UserException):
    def __init__(self, message: str, backup_id: str):
        super().__init__(message)
        self.backup_id = backup_id

    def __str__(self):
        return f'ERROR: On the [{self.backup_id}] backup\n' \
               f'Reason: {self.message}'


class ModuleException(BackupException):
    def __init__(self, message: str, backup_id: str, module: str):
        super().__init__(message, backup_id)
        self.module = module

    def __str__(self):
        return f'ERROR: On the [{self.backup_id}] backup\n' \
               f'On the [{self.module}] module\n' \
               f'Reason: {self.message}'


class ParameterException(ModuleException):
    def __init__(self, message: str, parameter: str, backup_id: str, module: str):
        super().__init__(message, backup_id, module)
        self.parameter = parameter

    def __str__(self):
        return f'ERROR: On the [{self.backup_id}] backup\n' \
               f'On the [{self.module}] module\n' \
               f'Validation error on parameter: {self.parameter}\n' \
               f'Reason: {self.message}'



# Copyright (c) 2015-2019, Activision Publishing, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

__tracebackhide__ = True


class DictMixin(object):
    """Dict assertions mixin."""

    def contains_key(self, *keys):
        """Asserts the val is a dict and contains the given key or keys.  Alias for :meth:`~assertpy.contains.ContainsMixin.contains`.

        Checks if the dict contains the given key or keys using ``in`` operator.

        Args:
            *keys: the key or keys expected to be contained

        Examples:
            Usage::

                assert_that({'a': 1, 'b': 2}).contains_key('a')
                assert_that({'a': 1, 'b': 2}).contains_key('a', 'b')

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val does **not** contain the key or keys
        """
        self._check_dict_like(self.val, check_values=False, check_getitem=False)
        return self.contains(*keys)

    def does_not_contain_key(self, *keys):
        """Asserts the val is a dict and does not contain the given key or keys.  Alias for :meth:`~assertpy.contains.ContainsMixin.does_not_contain`.

        Checks if the dict excludes the given key or keys using ``in`` operator.

        Args:
            *keys: the key or keys expected to be excluded

        Examples:
            Usage::

                assert_that({'a': 1, 'b': 2}).does_not_contain_key('x')
                assert_that({'a': 1, 'b': 2}).does_not_contain_key('x', 'y')

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val **does** contain the key or keys
        """
        self._check_dict_like(self.val, check_values=False, check_getitem=False)
        return self.does_not_contain(*keys)

    def contains_value(self, *values):
        """Asserts that val is a dict and contains the given value or values.

        Checks if the dict contains the given value or values in *any* key.

        Args:
            *values: the value or values expected to be contained

        Examples:
            Usage::

                assert_that({'a': 1, 'b': 2}).contains_value(1)
                assert_that({'a': 1, 'b': 2}).contains_value(1, 2)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val does **not** contain the value or values
        """
        self._check_dict_like(self.val, check_getitem=False)
        if len(values) == 0:
            raise ValueError('one or more value args must be given')
        missing = []
        for v in values:
            if v not in self.val.values():
                missing.append(v)
        if missing:
            self.error('Expected <%s> to contain values %s, but did not contain %s.' % (self.val, self._fmt_items(values), self._fmt_items(missing)))
        return self

    def does_not_contain_value(self, *values):
        """Asserts that val is a dict and does not contain the given value or values.

        Checks if the dict excludes the given value or values across *all* keys.

        Args:
            *values: the value or values expected to be excluded

        Examples:
            Usage::

                assert_that({'a': 1, 'b': 2}).does_not_contain_value(3)
                assert_that({'a': 1, 'b': 2}).does_not_contain_value(3, 4)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val **does** contain the value or values
        """
        self._check_dict_like(self.val, check_getitem=False)
        if len(values) == 0:
            raise ValueError('one or more value args must be given')
        else:
            found = []
            for v in values:
                if v in self.val.values():
                    found.append(v)
            if found:
                self.error('Expected <%s> to not contain values %s, but did contain %s.' % (self.val, self._fmt_items(values), self._fmt_items(found)))
        return self

    def contains_entry(self, *args, **kwargs):
        """Asserts that val is a dict and contains the given entry or entries.

        Checks if the dict contains the given key-value pair or pairs.

        Args:
            *args: the entry or entries expected to be contained (as ``{k: v}`` args)
            **kwargs: the entry or entries expected to be contained (as ``k=v`` kwargs)

        Examples:
            Usage::

                # using args
                assert_that({'a': 1, 'b': 2, 'c': 3}).contains_entry({'a': 1})
                assert_that({'a': 1, 'b': 2, 'c': 3}).contains_entry({'a': 1}, {'b': 2})
                assert_that({'a': 1, 'b': 2, 'c': 3}).contains_entry({'a': 1}, {'b': 2}, {'c': 3})

                # using kwargs
                assert_that({'a': 1, 'b': 2, 'c': 3}).contains_entry(a=1)
                assert_that({'a': 1, 'b': 2, 'c': 3}).contains_entry(a=1, b=2)
                assert_that({'a': 1, 'b': 2, 'c': 3}).contains_entry(a=1, b=2, c=3)

                # or args and kwargs
                assert_that({'a': 1, 'b': 2, 'c': 3}).contains_entry({'c': 3}, a=1, b=2)


        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val does **not** contain the entry or entries
        """
        self._check_dict_like(self.val, check_values=False)
        entries = list(args) + [{k: v} for k, v in kwargs.items()]
        if len(entries) == 0:
            raise ValueError('one or more entry args must be given')
        missing = []
        for e in entries:
            if type(e) is not dict:
                raise TypeError('given entry arg must be a dict')
            if len(e) != 1:
                raise ValueError('given entry args must contain exactly one key-value pair')
            k = next(iter(e))
            if k not in self.val:
                missing.append(e)  # bad key
            elif self.val[k] != e[k]:
                missing.append(e)  # bad val
        if missing:
            self.error('Expected <%s> to contain entries %s, but did not contain %s.' % (self.val, self._fmt_items(entries), self._fmt_items(missing)))
        return self

    def does_not_contain_entry(self, *args, **kwargs):
        """Asserts that val is a dict and does not contain the given entry or entries.

        Checks if the dict excludes the given key-value pair or pairs.

        Args:
            *args: the entry or entries expected to be excluded (as ``{k: v}`` args)
            **kwargs: the entry or entries expected to be excluded (as ``k=v`` kwargs)

        Examples:
            Usage::

                # using args
                assert_that({'a': 1, 'b': 2, 'c': 3}).does_not_contain_entry({'a': 2})
                assert_that({'a': 1, 'b': 2, 'c': 3}).does_not_contain_entry({'a': 2}, {'x': 4})

                # using kwargs
                assert_that({'a': 1, 'b': 2, 'c': 3}).does_not_contain_entry(a=2)
                assert_that({'a': 1, 'b': 2, 'c': 3}).does_not_contain_entry(a=2, x=4)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val **does** contain the entry or entries
        """
        self._check_dict_like(self.val, check_values=False)
        entries = list(args) + [{k: v} for k, v in kwargs.items()]
        if len(entries) == 0:
            raise ValueError('one or more entry args must be given')
        found = []
        for e in entries:
            if type(e) is not dict:
                raise TypeError('given entry arg must be a dict')
            if len(e) != 1:
                raise ValueError('given entry args must contain exactly one key-value pair')
            k = next(iter(e))
            if k in self.val and e[k] == self.val[k]:
                found.append(e)
        if found:
            self.error('Expected <%s> to not contain entries %s, but did contain %s.' % (self.val, self._fmt_items(entries), self._fmt_items(found)))
        return self



import os
import shutil
from pathlib import Path

from _pytest.fixtures import fixture

tests_path = Path(os.path.dirname(os.path.realpath(__file__)))


@fixture(autouse=True)
def change_test_dir(monkeypatch):
    """ Change working directory to test folder """
    monkeypatch.chdir(tests_path)


@fixture
def backup_folder():
    """ Create and remove a folder """
    backup_path = tests_path / 'backup'
    os.makedirs(backup_path, exist_ok=True)
    yield backup_path
    # Clear output folder
    shutil.rmtree(backup_path)


@fixture
def server_data_folder():
    """ Copy resources/testFolder to server_data_folder then remove it """
    server_data_path = tests_path / 'serverData'
    shutil.copytree(tests_path / 'resources' / 'testFolder', server_data_path)
    yield server_data_path
    # Clear output folder
    shutil.rmtree(server_data_path)



import locale
import logging
import os
from pathlib import Path

from assertpy import assert_that
from freezegun import freeze_time

from bashckup.bashckup import main

current_path = Path(os.path.dirname(os.path.realpath(__file__)))
tests_path = current_path / '..' / '..'
conf_path = tests_path / 'resources' / 'confs'
locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')

"""
Depends on TAR
"""


# FIXME this test has to be rewritten
@freeze_time('2023-07-10 15:02:10')
def test_tar_gz_verbose(caplog, backup_folder, server_data_folder):
    """
    GOAL: Test verbose mode (it has to be tested on every reader and post-backup)
    """
    caplog.set_level(logging.DEBUG)
    # Given
    config_file = conf_path / 'tar-gz.yml'
    # When
    return_code = main(['--verbose', 'backup', 'file', '--config-file', str(config_file)])

    # Then
    assert_that(return_code).is_equal_to(0)
    assert any(record.levelno == logging.DEBUG for record in caplog.records)


# FIXME this test has to be rewritten
@freeze_time('2023-07-10 15:02:10')
def test_tar_gz_quiet(caplog, backup_folder, server_data_folder):
    """
    GOAL: Test quiet mode
    """
    # Given
    config_file = conf_path / 'tar-gz.yml'
    # When
    return_code = main(['--quiet', 'backup', 'file', '--config-file', str(config_file)])

    # Then
    assert_that(return_code).is_equal_to(0)
    assert_that(caplog.records).is_empty()



# Copyright (c) 2015-2019, Activision Publishing, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import collections

if sys.version_info[0] == 3:
    str_types = (str,)
    Iterable = collections.abc.Iterable
else:
    str_types = (basestring,)
    Iterable = collections.Iterable

__tracebackhide__ = True


class ExtractingMixin(object):
    """Collection flattening mixin.

    It is often necessary to test collections of objects.  Use the ``extracting()`` helper to
    reduce the collection on a given attribute.  Reduce a list of objects::

        alice = Person('Alice', 'Alpha')
        bob = Person('Bob', 'Bravo')
        people = [alice, bob]

        assert_that(people).extracting('first_name').is_equal_to(['Alice', 'Bob'])
        assert_that(people).extracting('first_name').contains('Alice', 'Bob')
        assert_that(people).extracting('first_name').does_not_contain('Charlie')

    Additionally, the ``extracting()`` helper can accept a list of attributes to be extracted, and
    will flatten them into a list of tuples.  Reduce a list of objects on multiple attributes::

        assert_that(people).extracting('first_name', 'last_name').contains(('Alice', 'Alpha'), ('Bob', 'Bravo'))

    Also, ``extracting()`` works on not just attributes, but also properties, and even
    zero-argument methods.  Reduce a list of object on properties and zero-arg methods::

        assert_that(people).extracting('name').contains('Alice Alpha', 'Bob Bravo')
        assert_that(people).extracting('say_hello').contains('Hello, Alice!', 'Hello, Bob!')

    And ``extracting()`` even works on *dict-like* objects.  Reduce a list of dicts on key::

        alice = {'first_name': 'Alice', 'last_name': 'Alpha'}
        bob = {'first_name': 'Bob', 'last_name': 'Bravo'}
        people = [alice, bob]

        assert_that(people).extracting('first_name').contains('Alice', 'Bob')

    **Filtering**

    The ``extracting()`` helper can include a *filter* to keep only those items for which the given
    *filter* is truthy.  For example::

        users = [
            {'user': 'Alice', 'age': 36, 'active': True},
            {'user': 'Bob', 'age': 40, 'active': False},
            {'user': 'Charlie', 'age': 13, 'active': True}
        ]

        # filter the active users
        assert_that(users).extracting('user', filter='active').is_equal_to(['Alice', 'Charlie'])

    The *filter* can be a *dict-like* object and the extracted items are kept if and only if all
    corresponding key-value pairs are equal::

        assert_that(users).extracting('user', filter={'active': False}).is_equal_to(['Bob'])
        assert_that(users).extracting('user', filter={'age': 36, 'active': True}).is_equal_to(['Alice'])

    Or a *filter* can be any function (including an in-line ``lambda``) that accepts as its single
    argument each item in the collection, and the extracted items are kept if the function
    evaluates to ``True``::

        assert_that(users).extracting('user', filter=lambda x: x['age'] > 20)
            .is_equal_to(['Alice', 'Bob'])

    **Sorting**

    The ``extracting()`` helper can include a *sort* to enforce order on the extracted items.

    The *sort* can be the name of a key (or attribute, or property, or zero-argument method) and
    the extracted items are ordered by the corresponding values::

        assert_that(users).extracting('user', sort='age').is_equal_to(['Charlie', 'Alice', 'Bob'])

    The *sort* can be an ``iterable`` of names and the extracted items are ordered by
    corresponding value of the first name, ties are broken by the corresponding values of the
    second name, and so on::

        assert_that(users).extracting('user', sort=['active', 'age']).is_equal_to(['Bob', 'Charlie', 'Alice'])

    The *sort* can be any function (including an in-line ``lambda``) that accepts as its single
    argument each item in the collection, and the extracted items are ordered by the corresponding
    function return values::

        assert_that(users).extracting('user', sort=lambda x: -x['age']).is_equal_to(['Bob', 'Alice', 'Charlie'])
    """

    def extracting(self, *names, **kwargs):
        """Asserts that val is iterable, then extracts the named attributes, properties, or
        zero-arg methods into a list (or list of tuples if multiple names are given).

        Args:
            *names: the attribute to be extracted (or property or zero-arg method)
            **kwargs: see below

        Keyword Args:
            filter: extract only those items where filter is truthy
            sort: order the extracted items by the sort key

        Examples:
            Usage::

                alice = User('Alice', 20, True)
                bob = User('Bob', 30, False)
                charlie = User('Charlie', 10, True)
                users = [alice, bob, charlie]

                assert_that(users).extracting('user').contains('Alice', 'Bob', 'Charlie')

            Works with *dict-like* objects too::

                users = [
                    {'user': 'Alice', 'age': 20, 'active': True},
                    {'user': 'Bob', 'age': 30, 'active': False},
                    {'user': 'Charlie', 'age': 10, 'active': True}
                ]

                assert_that(people).extracting('user').contains('Alice', 'Bob', 'Charlie')

            Filter::

                assert_that(users).extracting('user', filter='active').is_equal_to(['Alice', 'Charlie'])

            Sort::

                assert_that(users).extracting('user', sort='age').is_equal_to(['Charlie', 'Alice', 'Bob'])

        Returns:
            AssertionBuilder: returns a new instance (now with the extracted list as the val) to chain to the next assertion
        """
        if not isinstance(self.val, Iterable):
            raise TypeError('val is not iterable')
        if isinstance(self.val, str_types):
            raise TypeError('val must not be string')
        if len(names) == 0:
            raise ValueError('one or more name args must be given')

        def _extract(x, name):
            if self._check_dict_like(x, check_values=False, return_as_bool=True):
                if name in x:
                    return x[name]
                else:
                    raise ValueError('item keys %s did not contain key <%s>' % (list(x.keys()), name))
            elif isinstance(x, Iterable):
                self._check_iterable(x, name='item')
                return x[name]
            elif hasattr(x, name):
                attr = getattr(x, name)
                if callable(attr):
                    try:
                        return attr()
                    except TypeError:
                        raise ValueError('val method <%s()> exists, but is not zero-arg method' % name)
                else:
                    return attr
            else:
                raise ValueError('val does not have property or zero-arg method <%s>' % name)

        def _filter(x):
            if 'filter' in kwargs:
                if isinstance(kwargs['filter'], str_types):
                    return bool(_extract(x, kwargs['filter']))
                elif self._check_dict_like(kwargs['filter'], check_values=False, return_as_bool=True):
                    for k in kwargs['filter']:
                        if isinstance(k, str_types):
                            if _extract(x, k) != kwargs['filter'][k]:
                                return False
                    return True
                elif callable(kwargs['filter']):
                    return kwargs['filter'](x)
                return False
            return True

        def _sort(x):
            if 'sort' in kwargs:
                if isinstance(kwargs['sort'], str_types):
                    return _extract(x, kwargs['sort'])
                elif isinstance(kwargs['sort'], Iterable):
                    items = []
                    for k in kwargs['sort']:
                        if isinstance(k, str_types):
                            items.append(_extract(x, k))
                    return tuple(items)
                elif callable(kwargs['sort']):
                    return kwargs['sort'](x)
            return 0

        extracted = []
        for i in sorted(self.val, key=lambda x: _sort(x)):
            if _filter(i):
                items = [_extract(i, name) for name in names]
                extracted.append(tuple(items) if len(items) > 1 else items[0])

        # chain on with _extracted_ list (don't chain to self!)
        return self.builder(extracted, self.description, self.kind)



# Copyright (c) 2015-2019, Activision Publishing, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import collections

if sys.version_info[0] == 3:
    Iterable = collections.abc.Iterable
else:
    Iterable = collections.Iterable

__tracebackhide__ = True


class CollectionMixin(object):
    """Collection assertions mixin."""

    def is_iterable(self):
        """Asserts that val is iterable collection.

        Examples:
            Usage::

                assert_that('foo').is_iterable()
                assert_that(['a', 'b']).is_iterable()
                assert_that((1, 2, 3)).is_iterable()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** iterable
        """
        if not isinstance(self.val, Iterable):
            self.error('Expected iterable, but was not.')
        return self

    def is_not_iterable(self):
        """Asserts that val is not iterable collection.

        Examples:
            Usage::

                assert_that(1).is_not_iterable()
                assert_that(123.4).is_not_iterable()
                assert_that(True).is_not_iterable()
                assert_that(None).is_not_iterable()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val **is** iterable
        """
        if isinstance(self.val, Iterable):
            self.error('Expected not iterable, but was.')
        return self

    def is_subset_of(self, *supersets):
        """Asserts that val is iterable and a subset of the given superset (or supersets).

        Args:
            *supersets: the expected superset (or supersets)

        Examples:
            Usage::

                assert_that('foo').is_subset_of('abcdefghijklmnopqrstuvwxyz')
                assert_that(['a', 'b']).is_subset_of(['a', 'b', 'c'])
                assert_that((1, 2, 3)).is_subset_of([1, 2, 3, 4])
                assert_that({'a': 1, 'b': 2}).is_subset_of({'a': 1, 'b': 2, 'c': 3})
                assert_that({'a', 'b'}).is_subset_of({'a', 'b', 'c'})

                # or multiple supersets (as comma-separated args)
                assert_that('aBc').is_subset_of('abc', 'ABC')
                assert_that((1, 2, 3)).is_subset_of([1, 3, 5], [2, 4, 6])

                assert_that({'a': 1, 'b': 2}).is_subset_of({'a': 1, 'c': 3})  # fails
                # Expected <{'a': 1, 'b': 2}> to be subset of <{'a': 1, 'c': 3}>, but <{'b': 2}> was missing.

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** subset of given superset (or supersets)
        """
        if not isinstance(self.val, Iterable):
            raise TypeError('val is not iterable')
        if len(supersets) == 0:
            raise ValueError('one or more superset args must be given')

        missing = []
        if hasattr(self.val, 'keys') and callable(getattr(self.val, 'keys')) and hasattr(self.val, '__getitem__'):
            # flatten superset dicts
            superdict = {}
            for l, j in enumerate(supersets):
                self._check_dict_like(j, check_values=False, name='arg #%d' % (l+1))
                for k in j.keys():
                    superdict.update({k: j[k]})

            for i in self.val.keys():
                if i not in superdict:
                    missing.append({i: self.val[i]})  # bad key
                elif self.val[i] != superdict[i]:
                    missing.append({i: self.val[i]})  # bad val
            if missing:
                self.error('Expected <%s> to be subset of %s, but %s %s missing.' % (
                    self.val, self._fmt_items(superdict), self._fmt_items(missing), 'was' if len(missing) == 1 else 'were'))
        else:
            # flatten supersets
            superset = set()
            for j in supersets:
                try:
                    for k in j:
                        superset.add(k)
                except Exception:
                    superset.add(j)

            for i in self.val:
                if i not in superset:
                    missing.append(i)
            if missing:
                self.error('Expected <%s> to be subset of %s, but %s %s missing.' % (
                    self.val, self._fmt_items(superset), self._fmt_items(missing), 'was' if len(missing) == 1 else 'were'))

        return self

    def is_sorted(self, key=lambda x: x, reverse=False):
        """Asserts that val is iterable and is sorted.

        Args:
            key (function): the one-arg function to extract the sort comparison key.  Defaults to
                ``lambda x: x`` to just compare items directly.
            reverse (bool): if ``True``, then comparison key is reversed.  Defaults to ``False``.

        Examples:
            Usage::

                assert_that(['a', 'b', 'c']).is_sorted()
                assert_that((1, 2, 3)).is_sorted()

                # with a key function
                assert_that('aBc').is_sorted(key=str.lower)

                # reverse order
                assert_that(['c', 'b', 'a']).is_sorted(reverse=True)
                assert_that((3, 2, 1)).is_sorted(reverse=True)

                assert_that((1, 2, 3, 4, -5, 6)).is_sorted()  # fails
                # Expected <(1, 2, 3, 4, -5, 6)> to be sorted, but subset <4, -5> at index 3 is not.

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** sorted
        """
        if not isinstance(self.val, Iterable):
            raise TypeError('val is not iterable')

        for i, x in enumerate(self.val):
            if i > 0:
                if reverse:
                    if key(x) > key(prev):
                        self.error('Expected <%s> to be sorted reverse, but subset %s at index %s is not.' % (self.val, self._fmt_items([prev, x]), i-1))
                else:
                    if key(x) < key(prev):
                        self.error('Expected <%s> to be sorted, but subset %s at index %s is not.' % (self.val, self._fmt_items([prev, x]), i-1))
            prev = x

        return self



import locale
import os
import re
import shutil
import subprocess
from pathlib import Path

from _pytest.fixtures import fixture
from assertpy import assert_that
from freezegun import freeze_time

from bashckup.bashckup import main

current_path = Path(os.path.dirname(os.path.realpath(__file__)))
tests_path = current_path / '..' / '..'
conf_path = tests_path / 'resources' / 'confs'
files_path = tests_path / 'resources' / 'mariadb'
locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')


@freeze_time('2023-07-10 15:02:10')
def test_restore_mariadb(backup_folder):
    """
    GOAL: Test restoration: all modifications have to be reverted
    """
    # Given
    config_file = conf_path / 'mariadb.yml'
    expected_backup_folder = backup_folder / 'mariadb'
    os.makedirs(expected_backup_folder)

    # Do some modifications
    process = subprocess.run(['mysql', '-D', 'test', '-e', "INSERT INTO `MOCK_DATA` (`id`, `app_name`, `app_version`, "
                                                           "`color_theme`) VALUES ('1001', 'bashckup-test', '0.1.0', "
                                                           "'blouge'); "
                                                           "UPDATE `MOCK_DATA` SET `app_version` = '0.0.1' "
                                                           "WHERE id = '42'"])
    shutil.copyfile(files_path / '2023-07-10T15:02:10-mariadb.sql',
                    expected_backup_folder / '2023-07-10T15:02:10-mariadb.sql')
    if process.returncode != 0:
        raise Exception("Unable run mysql command")

    # When
    return_code = main(['restore', 'file', '--config-file', str(config_file)])

    # Then
    assert_that(return_code).is_equal_to(0)

    # Remove 'comments'
    regex = r"/\*![0-9]{5} ?(?:[^*]*)\*/(?:;)?\n?"
    process = subprocess.run(['mysqldump', '--compact', '--skip-extended-insert', 'test'], capture_output=True,
                             shell=False, text=True)
    sql_dump = re.sub(regex, '', process.stdout, 0, re.MULTILINE)
    expected = (tests_path / '..' / '.github' / 'images' / 'database.sql').read_text().replace('COLLATE utf8mb4_bin ',
                                                                                               '')
    assert_that(sql_dump.replace('COLLATE utf8mb4_bin ', '')).is_equal_to(expected)



import locale
import logging
import os
import shutil
from pathlib import Path

import pytest
from assertpy import assert_that
from freezegun import freeze_time

from bashckup.bashckup import main

current_path = Path(os.path.dirname(os.path.realpath(__file__)))
tests_path = current_path / '..' / '..'
conf_path = tests_path / 'resources' / 'confs'
files_path = tests_path / 'resources' / 'tar'
locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')

"""
Depends on TAR
"""


@freeze_time('2023-07-10 15:02:10')
def test_output_file_permissions(backup_folder, server_data_folder):
    """
    Goal: Test permission of the output file
    """
    # Given
    config_file = conf_path / 'tar.yml'
    expected_backup_folder = backup_folder / 'tar'
    # When
    return_code = main(['backup', 'file', '--config-file', str(config_file)])

    # Then
    assert_that(return_code).is_equal_to(0)
    output = []
    with os.scandir(expected_backup_folder) as it:
        entry: os.DirEntry
        for entry in it:
            output.append({'file-name': entry.name})

    assert_that(output).contains_only({'file-name': '2023-07-10T15:02:10-tar.tar'})
    actual_permissions = oct(os.stat(expected_backup_folder / '2023-07-10T15:02:10-tar.tar').st_mode & 0o777)
    assert_that(actual_permissions).is_equal_to(oct(0o600))


@freeze_time('2023-07-10 15:02:10')
def test_output_file_file_already_exists(backup_folder, server_data_folder, caplog):
    """
    Goal: Test permission of the output file
    """
    # Given
    caplog.set_level(logging.ERROR)
    config_file = conf_path / 'tar.yml'
    expected_backup_folder = backup_folder / 'tar'
    os.mkdir(expected_backup_folder)
    with open(expected_backup_folder / '2023-07-10T15:02:10-tar.tar', 'w') as f:
        f.write('Already existing backup')
    # When
    with pytest.raises(SystemExit) as e:
        main(['backup', 'file', '--config-file', str(config_file)])

    # Then
    assert e.type == SystemExit
    assert e.value.code == 1
    assert_that(caplog.record_tuples).contains(
        ('root', logging.ERROR, 'ERROR: On the [tar] backup\n'
                                'On the [outputFile] module\n'
                                'Reason: File [backup/tar/2023-07-10T15:02:10-tar.tar] already exists'))


@freeze_time('2023-07-10 15:02:10')
def test_output_file_file_snar_permissions(backup_folder, server_data_folder):
    """
    GOAL: Test permission of snar file
    """
    # Given
    config_file = conf_path / 'tar-diff.yml'
    expected_backup_folder = backup_folder / 'tar-diff'
    # When
    return_code = main(['backup', 'file', '--config-file', str(config_file)])

    # Then
    assert_that(return_code).is_equal_to(0)
    output = []
    with os.scandir(expected_backup_folder) as it:
        entry: os.DirEntry
        for entry in it:
            output.append({'file-name': entry.name})

    assert_that(output).is_length(2)
    snar_file = list(filter(lambda f: f['file-name'].endswith('.snar'), output))[0]
    bck_file = list(filter(lambda f: f['file-name'].endswith('.tar'), output))[0]
    assert_that(snar_file['file-name']).is_equal_to('2023-07-10T15:02:10-tar-snap-w28.snar')
    assert_that(oct(os.stat(expected_backup_folder / snar_file['file-name']).st_mode & 0o777)).is_equal_to(oct(0o600))

    assert_that(bck_file['file-name']).is_equal_to('2023-07-10T15:02:10-tar-diff.tar')
    assert_that(oct(os.stat(expected_backup_folder / bck_file['file-name']).st_mode & 0o777)).is_equal_to(oct(0o600))



# Copyright (c) 2015-2019, Activision Publishing, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

__tracebackhide__ = True


class BaseMixin(object):
    """Base mixin."""

    def described_as(self, description):
        """Describes the assertion.  On failure, the description is included in the error message.

        This is not an assertion itself.  But if the any of the following chained assertions fail,
        the description will be included in addition to the regular error message.

        Args:
            description: the error message description

        Examples:
            Usage::

                assert_that(1).described_as('error msg desc').is_equal_to(2)  # fails
                # [error msg desc] Expected <1> to be equal to <2>, but was not.

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion
        """
        self.description = str(description)
        return self

    def is_equal_to(self, other, **kwargs):
        """Asserts that val is equal to other.

        Checks actual is equal to expected using the ``==`` operator. When val is *dict-like*,
        optionally ignore or include keys when checking equality.

        Args:
            other: the expected value
            **kwargs: see below

        Keyword Args:
            ignore: the dict key (or list of keys) to ignore
            include: the dict key (of list of keys) to include

        Examples:
            Usage::

                assert_that(1 + 2).is_equal_to(3)
                assert_that('foo').is_equal_to('foo')
                assert_that(123).is_equal_to(123)
                assert_that(123.4).is_equal_to(123.4)
                assert_that(['a', 'b']).is_equal_to(['a', 'b'])
                assert_that((1, 2, 3)).is_equal_to((1, 2, 3))
                assert_that({'a': 1, 'b': 2}).is_equal_to({'a': 1, 'b': 2})
                assert_that({'a', 'b'}).is_equal_to({'a', 'b'})

            When the val is *dict-like*, keys can optionally be *ignored* when checking equality::

                # ignore a single key
                assert_that({'a': 1, 'b': 2}).is_equal_to({'a': 1}, ignore='b')

                # ignore multiple keys
                assert_that({'a': 1, 'b': 2, 'c': 3}).is_equal_to({'a': 1}, ignore=['b', 'c'])

                # ignore nested keys
                assert_that({'a': {'b': 2, 'c': 3, 'd': 4}}).is_equal_to({'a': {'d': 4}}, ignore=[('a', 'b'), ('a', 'c')])

            When the val is *dict-like*, only certain keys can be *included* when checking equality::

                # include a single key
                assert_that({'a': 1, 'b': 2}).is_equal_to({'a': 1}, include='a')

                # include multiple keys
                assert_that({'a': 1, 'b': 2, 'c': 3}).is_equal_to({'a': 1, 'b': 2}, include=['a', 'b'])

            Failure produces a nice error message::

                assert_that(1).is_equal_to(2)  # fails
                # Expected <1> to be equal to <2>, but was not.

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if actual is **not** equal to expected

        Tip:
            Using :meth:`is_equal_to` with a ``float`` val is just asking for trouble. Instead, you'll
            always want to use *fuzzy* numeric assertions like :meth:`~assertpy.numeric.NumericMixin.is_close_to`
            or :meth:`~assertpy.numeric.NumericMixin.is_between`.

        See Also:
            :meth:`~assertpy.string.StringMixin.is_equal_to_ignoring_case` - for case-insensitive string equality
        """
        if self._check_dict_like(self.val, check_values=False, return_as_bool=True) and \
                self._check_dict_like(other, check_values=False, return_as_bool=True):
            if self._dict_not_equal(self.val, other, ignore=kwargs.get('ignore'), include=kwargs.get('include')):
                self._dict_err(self.val, other, ignore=kwargs.get('ignore'), include=kwargs.get('include'))
        else:
            if self.val != other:
                self.error('Expected <%s> to be equal to <%s>, but was not.' % (self.val, other))
        return self

    def is_not_equal_to(self, other):
        """Asserts that val is not equal to other.

        Checks actual is not equal to expected using the ``!=`` operator.

        Args:
            other: the expected value

        Examples:
            Usage::

                assert_that(1 + 2).is_not_equal_to(4)
                assert_that('foo').is_not_equal_to('bar')
                assert_that(123).is_not_equal_to(456)
                assert_that(123.4).is_not_equal_to(567.8)
                assert_that(['a', 'b']).is_not_equal_to(['c', 'd'])
                assert_that((1, 2, 3)).is_not_equal_to((1, 2, 4))
                assert_that({'a': 1, 'b': 2}).is_not_equal_to({'a': 1, 'b': 3})
                assert_that({'a', 'b'}).is_not_equal_to({'a', 'x'})

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if actual **is** equal to expected
        """
        if self.val == other:
            self.error('Expected <%s> to be not equal to <%s>, but was.' % (self.val, other))
        return self

    def is_same_as(self, other):
        """Asserts that val is identical to other.

        Checks actual is identical to expected using the ``is`` operator.

        Args:
            other: the expected value

        Examples:
            Basic types are identical::

                assert_that(1).is_same_as(1)
                assert_that('foo').is_same_as('foo')
                assert_that(123.4).is_same_as(123.4)

            As are immutables like ``tuple``::

                assert_that((1, 2, 3)).is_same_as((1, 2, 3))

            But mutable collections like ``list``, ``dict``, and ``set`` are not::

                # these all fail...
                assert_that(['a', 'b']).is_same_as(['a', 'b'])  # fails
                assert_that({'a': 1, 'b': 2}).is_same_as({'a': 1, 'b': 2})  # fails
                assert_that({'a', 'b'}).is_same_as({'a', 'b'})  # fails

            Unless they are the same object::

                x = {'a': 1, 'b': 2}
                y = x
                assert_that(x).is_same_as(y)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if actual is **not** identical to expected
        """
        if self.val is not other:
            self.error('Expected <%s> to be identical to <%s>, but was not.' % (self.val, other))
        return self

    def is_not_same_as(self, other):
        """Asserts that val is not identical to other.

        Checks actual is not identical to expected using the ``is`` operator.

        Args:
            other: the expected value

        Examples:
            Usage::

                assert_that(1).is_not_same_as(2)
                assert_that('foo').is_not_same_as('bar')
                assert_that(123.4).is_not_same_as(567.8)
                assert_that((1, 2, 3)).is_not_same_as((1, 2, 4))

                # mutable collections, even if equal, are not identical...
                assert_that(['a', 'b']).is_not_same_as(['a', 'b'])
                assert_that({'a': 1, 'b': 2}).is_not_same_as({'a': 1, 'b': 2})
                assert_that({'a', 'b'}).is_not_same_as({'a', 'b'})

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if actual **is** identical to expected
        """
        if self.val is other:
            self.error('Expected <%s> to be not identical to <%s>, but was.' % (self.val, other))
        return self

    def is_true(self):
        """Asserts that val is true.

        Examples:
            Usage::

                assert_that(True).is_true()
                assert_that(1).is_true()
                assert_that('foo').is_true()
                assert_that(1.0).is_true()
                assert_that(['a', 'b']).is_true()
                assert_that((1, 2, 3)).is_true()
                assert_that({'a': 1, 'b': 2}).is_true()
                assert_that({'a', 'b'}).is_true()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val **is** false
        """
        if not self.val:
            self.error('Expected <True>, but was not.')
        return self

    def is_false(self):
        """Asserts that val is false.

        Examples:
            Usage::

                assert_that(False).is_false()
                assert_that(0).is_false()
                assert_that('').is_false()
                assert_that(0.0).is_false()
                assert_that([]).is_false()
                assert_that(()).is_false()
                assert_that({}).is_false()
                assert_that(set()).is_false()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val **is** true
        """
        if self.val:
            self.error('Expected <False>, but was not.')
        return self

    def is_none(self):
        """Asserts that val is none.

        Examples:
            Usage::

                assert_that(None).is_none()
                assert_that(print('hello world')).is_none()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** none
        """
        if self.val is not None:
            self.error('Expected <%s> to be <None>, but was not.' % self.val)
        return self

    def is_not_none(self):
        """Asserts that val is not none.

        Examples:
            Usage::

                assert_that(0).is_not_none()
                assert_that('foo').is_not_none()
                assert_that(False).is_not_none()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val **is** none
        """
        if self.val is None:
            self.error('Expected not <None>, but was.')
        return self

    def _type(self, val):
        if hasattr(val, '__name__'):
            return val.__name__
        elif hasattr(val, '__class__'):
            return val.__class__.__name__
        return 'unknown'

    def is_type_of(self, some_type):
        """Asserts that val is of the given type.

        Args:
            some_type (type): the expected type

        Examples:
            Usage::

                assert_that(1).is_type_of(int)
                assert_that('foo').is_type_of(str)
                assert_that(123.4).is_type_of(float)
                assert_that(['a', 'b']).is_type_of(list)
                assert_that((1, 2, 3)).is_type_of(tuple)
                assert_that({'a': 1, 'b': 2}).is_type_of(dict)
                assert_that({'a', 'b'}).is_type_of(set)
                assert_that(True).is_type_of(bool)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** of the given type
        """
        if type(some_type) is not type and not issubclass(type(some_type), type):
            raise TypeError('given arg must be a type')
        if type(self.val) is not some_type:
            t = self._type(self.val)
            self.error('Expected <%s:%s> to be of type <%s>, but was not.' % (self.val, t, some_type.__name__))
        return self

    def is_instance_of(self, some_class):
        """Asserts that val is an instance of the given class.

        Args:
            some_class: the expected class

        Examples:
            Usage::

                assert_that(1).is_instance_of(int)
                assert_that('foo').is_instance_of(str)
                assert_that(123.4).is_instance_of(float)
                assert_that(['a', 'b']).is_instance_of(list)
                assert_that((1, 2, 3)).is_instance_of(tuple)
                assert_that({'a': 1, 'b': 2}).is_instance_of(dict)
                assert_that({'a', 'b'}).is_instance_of(set)
                assert_that(True).is_instance_of(bool)

            With a user-defined class::

                class Foo: pass
                f = Foo()
                assert_that(f).is_instance_of(Foo)
                assert_that(f).is_instance_of(object)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** an instance of the given class
        """
        try:
            if not isinstance(self.val, some_class):
                t = self._type(self.val)
                self.error('Expected <%s:%s> to be instance of class <%s>, but was not.' % (self.val, t, some_class.__name__))
        except TypeError:
            raise TypeError('given arg must be a class')
        return self

    def is_length(self, length):
        """Asserts that val is the given length.

        Checks val is the given length using the ``len()`` built-in.

        Args:
            length (int): the expected length

        Examples:
            Usage::

                assert_that('foo').is_length(3)
                assert_that(['a', 'b']).is_length(2)
                assert_that((1, 2, 3)).is_length(3)
                assert_that({'a': 1, 'b': 2}).is_length(2)
                assert_that({'a', 'b'}).is_length(2)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** the given length
        """
        if type(length) is not int:
            raise TypeError('given arg must be an int')
        if length < 0:
            raise ValueError('given arg must be a positive int')
        if len(self.val) != length:
            self.error('Expected <%s> to be of length <%d>, but was <%d>.' % (self.val, length, len(self.val)))
        return self



# Copyright (c) 2015-2019, Activision Publishing, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import sys
import datetime
import inspect
import json

__tracebackhide__ = True


class SnapshotMixin(object):
    """Snapshot mixin.

    Take a snapshot of a python data structure, store it on disk in JSON format, and automatically
    compare the latest data to the stored data on every test run.

    Functional testing (which snapshot testing falls under) is very much blackbox testing.  When
    something goes wrong, it's hard to pinpoint the issue, because functional tests typically
    provide minimal *isolation* as compared to unit tests.  On the plus side, snapshots typically
    do provide enormous *leverage* as a few well-placed snapshot tests can strongly verify that an
    application is working.  Similar coverage would otherwise require dozens if not hundreds of
    unit tests.

    **On-disk Format**

    Snapshots are stored in a readable JSON format.  For example::

        assert_that({'a': 1, 'b': 2, 'c': 3}).snapshot()

    Would be stored as::

        {
            "a": 1,
            "b": 2,
            "c": 3
        }

    The JSON formatting support most python data structures (dict, list, object, etc), but not custom
    binary data.

    **Updating**

    It's easy to update your snapshots...just delete them all and re-run the test suite to regenerate all snapshots.

    Note:
        Snapshots require Python 3.x
    """

    def snapshot(self, id=None, path='__snapshots'):
        """Asserts that val is identical to the on-disk snapshot stored previously.

        On the first run of a test before the snapshot file has been saved, a snapshot is created,
        stored to disk, and the test *always* passes.  But on all subsequent runs, val is compared
        to the on-disk snapshot, and the test fails if they don't match.

        Snapshot artifacts are stored in the ``__snapshots`` directory by default, and should be
        committed to source control alongside any code changes.

        Snapshots are identified by test filename plus line number by default.

        Args:
            id: the item or items expected to be contained
            path: the item or items expected to be contained

        Examples:
            Usage::

                assert_that(None).snapshot()
                assert_that(True).snapshot()
                assert_that(1).snapshot()
                assert_that(123.4).snapshot()
                assert_that('foo').snapshot()
                assert_that([1, 2, 3]).snapshot()
                assert_that({'a': 1, 'b': 2, 'c': 3}).snapshot()
                assert_that({'a', 'b', 'c'}).snapshot()
                assert_that(1 + 2j).snapshot()
                assert_that(someobj).snapshot()

            By default, snapshots are identified by test filename plus line number.  Alternately, you can specify a custom identifier using the ``id`` arg::

                assert_that({'a': 1, 'b': 2, 'c': 3}).snapshot(id='foo-id')


            By default, snapshots are stored in the ``__snapshots`` directory.  Alternately, you can specify a custom path using the ``path`` arg::

                assert_that({'a': 1, 'b': 2, 'c': 3}).snapshot(path='my-custom-folder')

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val does **not** equal to on-disk snapshot
        """
        if sys.version_info[0] < 3:
            raise NotImplementedError('snapshot testing requires Python 3')

        class _Encoder(json.JSONEncoder):
            def default(self, o):
                if isinstance(o, set):
                    return {'__type__': 'set', '__data__': list(o)}
                elif isinstance(o, complex):
                    return {'__type__': 'complex', '__data__': [o.real, o.imag]}
                elif isinstance(o, datetime.datetime):
                    return {'__type__': 'datetime', '__data__': o.strftime('%Y-%m-%d %H:%M:%S')}
                elif '__dict__' in dir(o) and type(o) is not type:
                    return {
                        '__type__': 'instance',
                        '__class__': o.__class__.__name__,
                        '__module__': o.__class__.__module__,
                        '__data__': o.__dict__
                    }
                return json.JSONEncoder.default(self, o)

        class _Decoder(json.JSONDecoder):
            def __init__(self):
                json.JSONDecoder.__init__(self, object_hook=self.object_hook)

            def object_hook(self, d):
                if '__type__' in d and '__data__' in d:
                    if d['__type__'] == 'set':
                        return set(d['__data__'])
                    elif d['__type__'] == 'complex':
                        return complex(d['__data__'][0], d['__data__'][1])
                    elif d['__type__'] == 'datetime':
                        return datetime.datetime.strptime(d['__data__'], '%Y-%m-%d %H:%M:%S')
                    elif d['__type__'] == 'instance':
                        mod = __import__(d['__module__'], fromlist=[d['__class__']])
                        klass = getattr(mod, d['__class__'])
                        inst = klass.__new__(klass)
                        inst.__dict__ = d['__data__']
                        return inst
                return d

        def _save(name, val):
            with open(name, 'w') as fp:
                json.dump(val, fp, indent=2, separators=(',', ': '), sort_keys=True, cls=_Encoder)

        def _load(name):
            with open(name, 'r') as fp:
                return json.load(fp, cls=_Decoder)

        def _name(path, name):
            try:
                return os.path.join(path, 'snap-%s.json' % name.replace(' ', '_').lower())
            except Exception:
                raise ValueError('failed to create snapshot filename, either bad path or bad name')

        if id:
            # custom id
            snapname = _name(path, id)
        else:
            # make id from filename and line number
            f = inspect.currentframe()
            fpath = os.path.basename(f.f_back.f_code.co_filename)
            fname = os.path.splitext(fpath)[0]
            lineno = str(f.f_back.f_lineno)
            snapname = _name(path, fname)

        if not os.path.exists(path):
            os.makedirs(path)

        if os.path.isfile(snapname):
            # snap exists, so load
            snap = _load(snapname)

            if id:
                # custom id, so test
                return self.is_equal_to(snap)
            else:
                if lineno in snap:
                    # found sub-snap, so test
                    return self.is_equal_to(snap[lineno])
                else:
                    # lineno not in snap, so create sub-snap and pass
                    snap[lineno] = self.val
                    _save(snapname, snap)
        else:
            # no snap, so create and pass
            _save(snapname, self.val if id else {lineno: self.val})

        return self



import argparse
import ast
import datetime
import logging
import os
import sys
import time

import yaml
from jsonschema import validate
from jsonschema.exceptions import ValidationError
from yaml import Loader
from yaml.loader import SafeLoader

from bashckup.actuators.actuators_factories import ActuatorFactory
from bashckup.actuators.exceptions import UserException, RunningException

yaml_schema = """
type: array
items:
  type: object
  properties:
    name:
      type: string
    id:
      type: string
      pattern: '[a-zA-Z0-9]+'
    reader:
      type: object
      minProperties: 1
      maxProperties: 1
      patternProperties:
        '.*':
          type: object
          properties:
            args:
              type: object
              additionalProperties: true
          additionalProperties: false
      additionalProperties: false
    transformers:
      type: array
      items:
        type:
          - object
          - string
        minProperties: 1
        maxProperties: 1
        patternProperties:
          '.*':
            type: object
            properties:
              args:
                type: object
                additionalProperties: true
            additionalProperties: false
        additionalProperties: false
    writer:
      type: object
      minProperties: 1
      maxProperties: 1
      patternProperties:
        '.*':
          type: object
          properties:
            args:
              type: object
              additionalProperties: true
          additionalProperties: false
      additionalProperties: false
    post-backup:
      type: array
      items:
        type:
          - object
          - string
        minProperties: 1
        maxProperties: 1
        patternProperties:
          '.*':
            type: object
            properties:
              args:
                type: object
                additionalProperties: true
            additionalProperties: false
        additionalProperties: false
  required:
    - reader
    - writer
    - id
"""


class KeyValue(argparse.Action):
    # Constructor calling
    def __call__(self, parser, namespace,
                 values, option_string=None):
        setattr(namespace, self.dest, dict())

        for value in values:
            # split it into key and value
            key, value = value.split('=')
            # assign into dictionary
            getattr(namespace, self.dest)[key] = ast.literal_eval(value)


class AppendKeyValue(argparse.Action):
    noOperation = 'nop'

    # Constructor calling
    def __call__(self, parser, namespace,
                 values, option_string=None):
        items = getattr(namespace, self.dest, None)
        if items is None:
            items = []
        setattr(namespace, self.dest, items)

        dictionary = dict()
        if values != [self.noOperation]:
            for value in values:
                # split it into key and value
                key, value = value.split('=')
                # assign into dictionary
                dictionary[key] = ast.literal_eval(value)
        getattr(namespace, self.dest).append(dictionary)


def read_config_from_file(config_file: argparse.FileType):
    with config_file as f:
        config = yaml.load(f, Loader=SafeLoader)
    try:
        validate(config, yaml.load(yaml_schema, Loader=Loader))
    except ValidationError as e:
        raise UserException('Validation error on config file\n'
                            f'On property {e.json_path}\n'
                            f'Error description: {e.message}\n'
                            f'''Helper: {e.schema.get('description')}''') from None
    ids = list(map(lambda c: c['id'], config))
    if len(ids) != len(config):
        raise UserException('Backup id must be unique')
    return config


def read_config_from_cli(parameters: argparse.Namespace):
    result = []
    backup = {'name': 'Backup from CLI', 'id': '.',  # Id = . for file path, it prevents to add another folder
              'reader': {'module': parameters.reader_module, 'args': parameters.reader_args},
              'transformers': [],
              'writer': {'module': parameters.writer_module, 'args': parameters.writer_args},
              'post-backup': []}
    if parameters.transformer_module is not None or parameters.transformer_args is not None:
        if parameters.transformer_module is None or parameters.transformer_args is None or len(
                parameters.transformer_module) != len(parameters.transformer_args):
            raise UserException("Number of transformers doesn't match number of transformer's arguments!")
        for i, v in enumerate(parameters.transformer_module):
            m = parameters.transformer_module[i]
            args = parameters.transformer_args[i]
            module = {m: {'args': args}}
            backup['transformers'].append(module)
    if parameters.post_backup_module is not None or parameters.post_backup_args is not None:
        if parameters.post_backup_module is None or parameters.post_backup_args is None or len(
                parameters.post_backup_module) != len(parameters.post_backup_args):
            raise UserException("Number of post-backup doesn't match number of post-backup's arguments!\n")
        for i, v in enumerate(parameters.post_backup_module):
            m = parameters.post_backup_module[i]
            args = parameters.post_backup_args[i]
            module = {m: {'args': args}}
            backup['post-backup'].append(module)
    result.append(backup)
    return result


def prepare(global_parameters: dict, configurations: dict) -> dict:
    result = {}
    for current_backup in configurations:
        metadata = {}
        modules = {}

        global_context = {**global_parameters, **{'backup-id': current_backup['id']}}

        reader = ActuatorFactory().build_reader(global_context, current_backup['reader'], metadata)
        modules.update({'reader': reader})
        metadata.update(reader.prepare_module())

        if current_backup.get('transformers') is not None:
            for transformer in current_backup['transformers']:
                trans = ActuatorFactory().build_transformer(global_context, transformer, metadata)
                transformers = modules.get('transformers', [])
                transformers.append(trans)
                modules.update({'transformers': transformers})
                metadata.update(trans.prepare_module())

        writer = ActuatorFactory().build_writer(global_context, current_backup['writer'], metadata)
        modules.update({'writer': writer})
        metadata.update(writer.prepare_module())

        if current_backup.get('post-backup') is not None:
            for post_backup in current_backup['post-backup']:
                post_bck = ActuatorFactory().build_post_backup(global_context, post_backup, metadata)
                post_backups = modules.get('post-backup', [])
                post_backups.append(post_bck)
                modules.update({'post-backup': post_backups})
                metadata.update(post_bck.prepare_module())

        result.update({current_backup['id']: {'modules': modules, 'metadata': metadata}})

    return result


def run_backup_plans(global_parameters: dict, backup_plans: dict) -> bool:
    """
    :returns: True if no errors appear during the backup, otherwise False.
    """
    error = False
    for (backup_id, backup_plan) in backup_plans.items():
        try:
            logging.info('=== Backup %s ===', backup_id)

            #
            # Backup
            #
            if global_parameters['dry-run'] is False:
                processes = []  # Store all processes to be able to retrieve errors
                try:
                    logging.info('= Run reader %s =', backup_plan['modules']['reader'].module_name())
                    processes.append(backup_plan['modules']['reader'].generate_backup_process())
                    if backup_plan['modules'].get('transformers') is not None:
                        for transformer in backup_plan['modules']['transformers']:
                            logging.info('= Run transformer %s =', transformer.module_name())
                            previous_process = processes[-1]
                            processes.append(transformer.generate_backup_process(previous_process.stdout))
                            previous_process.stdout.close()  # Allow previous process to receive a SIGPIPE

                    previous_process = processes[-1]
                    logging.info('= Run writer %s =', backup_plan['modules']['writer'].module_name())
                    processes.append(backup_plan['modules']['writer'].generate_backup_process(previous_process.stdout))
                    previous_process.stdout.close()  # Allow previous process to receive a SIGPIPE
                finally:
                    for process in processes:
                        return_code = process.wait()
                        if return_code != 0:
                            error = True
                            logging.error('ERROR: Error during execution of backup\n'
                                          f'Command output: {process.stderr.read().decode(sys.getdefaultencoding())}\n'
                                          f'''Command executed: {' '.join(process.args)}'''
                                          f'Error code: {return_code}')
                        else:
                            stderr = process.stderr.read().decode(sys.getdefaultencoding())
                            if stderr != '':
                                logging.debug(stderr)
                        if process.stderr is not None:
                            process.stderr.close()
                        if process.stdout is not None:
                            process.stdout.close()
            else:  # Dry run
                cmd = []
                cmd.extend(backup_plan['modules']['reader'].generate_dry_run_backup_cmd())
                if backup_plan['modules'].get('transformers') is not None:
                    for transformer in backup_plan['modules']['transformers']:
                        cmd.append('|')
                        cmd.extend(transformer.generate_dry_run_backup_cmd())

                cmd.append('|')
                cmd.extend(backup_plan['modules']['writer'].generate_dry_run_backup_cmd())
                logging.info(f'''Command [{' '.join(cmd)}] would have been ran.''')
            #
            # Post backup
            #
            if backup_plan['modules'].get('post-backup') is not None:
                logging.info('== Post backup ==')
                for post_backup in backup_plan['modules']['post-backup']:
                    logging.info('= Run post backup %s =', post_backup.module_name())
                    # TODO manage return code and errors
                    post_backup.run_backup()
        except (UserException, RunningException) as e:
            error = True
            logging.error(str(e))
    return error


def run_restoration_plans(global_parameters: dict, backup_plans: dict) -> bool:
    """
    :returns: True if no errors appear during the restoration, otherwise False.
    """
    error = False
    for (backup_id, backup_plan) in backup_plans.items():
        try:
            logging.info('=== Restoration backup %s ===', backup_id)

            #
            # Post backup
            #
            if backup_plan['modules'].get('post-backup') is not None:
                logging.info('== Post backup ==')
                for post_backup in backup_plan['modules']['post-backup'].reverse():
                    logging.info('= Run post backup %s =', post_backup.module_name())
                    # TODO manage return code and errors
                    post_backup.run_restore()

            #
            # Backup
            #
            if global_parameters['dry-run'] is False:
                processes = []  # Store all processes to be able to retrieve errors
                try:
                    logging.info('= Run writer %s =', backup_plan['modules']['writer'].module_name())
                    processes.append(backup_plan['modules']['writer'].generate_restore_process())

                    if backup_plan['modules'].get('transformers') is not None:
                        for transformer in backup_plan['modules']['transformers'].reverse():
                            logging.info('= Run transformer %s =', transformer.module_name())
                            previous_process = processes[-1]
                            processes.append(transformer.generate_restore_process(previous_process.stdout))
                            previous_process.stdout.close()  # Allow previous process to receive a SIGPIPE

                    previous_process = processes[-1]
                    logging.info('= Run reader %s =', backup_plan['modules']['reader'].module_name())
                    processes.append(backup_plan['modules']['reader'].generate_restore_process(previous_process.stdout))
                    previous_process.stdout.close()  # Allow previous process to receive a SIGPIPE

                finally:
                    for process in processes:
                        return_code = process.wait()
                        if return_code != 0:
                            error = True
                            logging.error('ERROR: Error during execution of backup\n'
                                          f'Command output: {process.stderr.read().decode(sys.getdefaultencoding())}\n'
                                          f'''Command executed: {' '.join(process.args)}'''
                                          f'Error code: {return_code}')
                        else:
                            stderr = process.stderr.read().decode(sys.getdefaultencoding())
                            if stderr != '':
                                logging.debug(stderr)
                        if process.stderr is not None:
                            process.stderr.close()
                        if process.stdout is not None:
                            process.stdout.close()
            else:  # Dry run
                cmd = []
                cmd.extend(backup_plan['modules']['writer'].generate_dry_run_restore_cmd())
                if backup_plan['modules'].get('transformers') is not None:
                    for transformer in backup_plan['modules']['transformers'].reverse():
                        cmd.append('|')
                        cmd.extend(transformer.generate_dry_run_restore_cmd())
                cmd.append('|')
                cmd.extend(backup_plan['modules']['reader'].generate_dry_run_restore_cmd())

                logging.info(f'''Command [{' '.join(cmd)}] would have been ran.''')
        except (UserException, RunningException) as e:
            error = True
            logging.error(str(e))
    return error


def extract_cli_parameters(args):
    args_parser = argparse.ArgumentParser(prog='Backuper', description='Command line interface to backup everything!')
    args_parser.add_argument('--dry-run', action='store_true',
                             help='Run program in dry mode, nothing is done on the system')
    args_parser.add_argument('--verbose', action='store_true',
                             help='Run program in verbose mode to have all information printed on stdout. '
                                  'Cannot be used together with --quiet')
    args_parser.add_argument('--quiet', action='store_true',
                             help='Never print into stdout, only warning and errors will be printed in stderr. '
                                  'Cannot be used together with --verbose')

    sub_parser = args_parser.add_subparsers(title='Mode', dest='mode', required=True,
                                            description='Backup or restore')
    backup_parser = sub_parser.add_parser('backup', help='Get config by CLI arguments')
    restore_parser = sub_parser.add_parser('restore', help='Get config from a YAML file')

    for parse in [backup_parser, restore_parser]:
        sub_parser = parse.add_subparsers(title='Config mode', dest='config_mode', required=True,
                                          description='How backup rules are defined')
        cli_parser = sub_parser.add_parser('cli', help='Get config by CLI arguments')
        file_parser = sub_parser.add_parser('file', help='Get config from a YAML file')
        file_parser.add_argument('--config-file', type=argparse.FileType('r'), required=True,
                                 help='Config wile to read')
        cli_parser.add_argument('--reader-module', choices=ActuatorFactory.reader_module_name(), required=True,
                                help='Module used to generate source backup')
        cli_parser.add_argument('--reader-args', action=KeyValue, nargs='*', help='Arguments for the reader module')
        cli_parser.add_argument('--transformer-module', choices=ActuatorFactory.transformer_module_name(),
                                action='append',
                                help='Module to transform backup, order may be used to create transformation chain')
        cli_parser.add_argument('--transformer-args', action=AppendKeyValue, nargs='*',
                                help='Arguments for the transform module. Number of args must match number of '
                                     'transforms, you can use nop as empty arg')
        cli_parser.add_argument('--writer-module', choices=ActuatorFactory.writer_module_name(), required=True,
                                help='Module used to generate source backup')
        cli_parser.add_argument('--writer-args', action=KeyValue, nargs='*', help='Arguments for the writer module')
        cli_parser.add_argument('--post-backup-module', choices=ActuatorFactory.post_backup_module_name(),
                                action='append',
                                help='Module to some tasks after backup is done')
        cli_parser.add_argument('--post-backup-args', action=AppendKeyValue, nargs='*',
                                help='Arguments for the post-backup module. Number of args must match number of '
                                     'post-backup, you can use nop as empty arg')
    parameters = args_parser.parse_args(args)

    # Checks some basics stuff
    if parameters.verbose is True and parameters.quiet is True:
        raise UserException("--verbose and --quiet cannot present together, choose one.")

    return parameters


def configure_logging(parameters):
    error_stream_handler = logging.StreamHandler(sys.stderr)
    error_stream_handler.setLevel(logging.WARNING)
    handlers = [error_stream_handler]
    if parameters.quiet is False:
        msg_stream_handler = logging.StreamHandler(sys.stdout)
        if parameters.verbose is True:
            msg_stream_handler.setLevel(logging.DEBUG)
        else:
            msg_stream_handler.setLevel(logging.INFO)
        handlers.append(msg_stream_handler)
    logging.basicConfig(format='%(levelname)s - %(message)s', handlers=handlers, level=logging.DEBUG)


def main(args=None):
    # Default rights to prevent backups to be visible for everyone
    os.umask(0o077)
    starting_time = time.time()
    try:
        parameters = extract_cli_parameters(args)

        configure_logging(parameters)
        global_parameters = {'dry-run': parameters.dry_run, 'verbose': parameters.verbose,
                             'backup': parameters.mode == 'backup'}

        if parameters.config_mode == 'file':
            configurations = read_config_from_file(parameters.config_file)
        elif parameters.config_mode == 'cli':
            configurations = read_config_from_cli(parameters)
        else:
            raise ValueError(f'Config mode {parameters.config_mode} is not handled')

        backup_plans = prepare(global_parameters, configurations)

        if global_parameters['backup']:
            have_error = run_backup_plans(global_parameters, backup_plans)
        else:
            have_error = run_restoration_plans(global_parameters, backup_plans)
        if have_error is True:
            exit(1)
        return 0
    except Exception as e:
        logging.error(str(e))
        exit(1)
    finally:
        if logging.getLogger().isEnabledFor(logging.INFO):
            time_elapsed = datetime.timedelta(seconds=time.time() - starting_time)
            logging.info('Backups took %s', str(time_elapsed))


if __name__ == "__main__":
    main()



# Copyright (c) 2015-2019, Activision Publishing, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import collections

if sys.version_info[0] == 3:
    Iterable = collections.abc.Iterable
else:
    Iterable = collections.Iterable

__tracebackhide__ = True


class ExceptionMixin(object):
    """Expected exception mixin."""

    def raises(self, ex):
        """Asserts that val is callable and set the expected exception.

        Just sets the expected exception, but never calls val, and therefore never failes. You must
        chain to :meth:`~when_called_with` to invoke ``val()``.

        Args:
            ex: the expected exception

        Examples:
            Usage::

                assert_that(some_func).raises(RuntimeError).when_called_with('foo')

        Returns:
            AssertionBuilder: returns a new instance (now with the given expected exception) to chain to the next assertion
        """
        if not callable(self.val):
            raise TypeError('val must be callable')
        if not issubclass(ex, BaseException):
            raise TypeError('given arg must be exception')

        # chain on with ex as the expected exception
        return self.builder(self.val, self.description, self.kind, ex)

    def when_called_with(self, *some_args, **some_kwargs):
        """Asserts that val, when invoked with the given args and kwargs, raises the expected exception.

        Invokes ``val()`` with the given args and kwargs.  You must first set the expected
        exception with :meth:`~raises`.

        Args:
            *some_args: the args to call ``val()``
            **some_kwargs: the kwargs to call ``val()``

        Examples:
            Usage::

                def some_func(a):
                    raise RuntimeError('some error!')

                assert_that(some_func).raises(RuntimeError).when_called_with('foo')

        Returns:
            AssertionBuilder: returns a new instance (now with the captured exception error message as the val) to chain to the next assertion

        Raises:
            AssertionError: if val does **not** raise the expected exception
            TypeError: if expected exception not set via :meth:`raises`
        """
        if not self.expected:
            raise TypeError('expected exception not set, raises() must be called first')
        try:
            self.val(*some_args, **some_kwargs)
        except BaseException as e:
            if issubclass(type(e), self.expected):
                # chain on with error message
                return self.builder(str(e), self.description, self.kind)
            else:
                # got exception, but wrong type, so raise
                self.error('Expected <%s> to raise <%s> when called with (%s), but raised <%s>.' % (
                    self.val.__name__,
                    self.expected.__name__,
                    self._fmt_args_kwargs(*some_args, **some_kwargs),
                    type(e).__name__))

        # didn't fail as expected, so raise
        self.error('Expected <%s> to raise <%s> when called with (%s).' % (
            self.val.__name__,
            self.expected.__name__,
            self._fmt_args_kwargs(*some_args, **some_kwargs)))



import locale
import logging
import os
import shutil
from pathlib import Path

from _pytest.fixtures import fixture
from assertpy import assert_that
from freezegun import freeze_time

from bashckup.bashckup import main

current_path = Path(os.path.dirname(os.path.realpath(__file__)))
tests_path = current_path / '..' / '..'
conf_path = tests_path / 'resources' / 'confs'
locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')

"""
Depends on TAR
"""


@freeze_time('2023-07-09 15:02:10')
def test_tar_clean(backup_folder, server_data_folder):
    """
    Goal: Test clean folder, file older than 2 days has to be removed
    """
    # Given
    config_file = conf_path / 'tar-clean.yml'
    expected_backup_folder = backup_folder / 'tar-clean'
    generate_file(expected_backup_folder, '2023-07-08T15:02:10-tar.tar')  # Expected to be kept (Diff 1 day)
    generate_file(expected_backup_folder,
                  '2023-07-08T15:02:11-tar.tar')  # Expected to be kept, less than 2 days (2 days minus 1 s)
    generate_file(expected_backup_folder, '2023-07-07T15:02:10-tar.tar')  # Expected to be removed exactly 2 days
    generate_file(expected_backup_folder, '2023-07-06T15:02:10-tar.tar')  # Expected to be removed, older than 3 days
    # When
    return_code = main(['backup', 'file', '--config-file', str(config_file)])

    # Then
    assert_that(return_code).is_equal_to(0)
    output = []
    with os.scandir(expected_backup_folder) as it:
        entry: os.DirEntry
        for entry in it:
            output.append({'file-name': entry.name, 'size': entry.stat().st_size})

    assert_that(output).contains_only({'file-name': '2023-07-09T15:02:10-tar-clean.tar', 'size': 10240},
                                      {'file-name': '2023-07-08T15:02:10-tar.tar', 'size': 7},
                                      {'file-name': '2023-07-08T15:02:11-tar.tar', 'size': 7})


@freeze_time('2023-07-09 15:02:10')
def test_tar_incremental_clean_retention_less_than_level0_frequency(caplog, backup_folder, server_data_folder):
    """
    Goal: Test clean folder with a retention for 'cleanFolder' lower than the incremental backup interval of tar
    3th of July is the first day of the current week
    """
    caplog.set_level(logging.WARNING)
    # Given
    config_file = conf_path / 'tar-diff-clean.yml'
    expected_backup_folder = backup_folder / 'tar-diff-clean'
    generate_file(expected_backup_folder, '2023-07-08T15:02:10-tar.tar')  # Expected to be kept (Diff 1 day)
    generate_file(expected_backup_folder,
                  '2023-07-07T15:02:11-tar.tar')  # Expected to be kept, less than 2 days (2 days minus 1 s)
    generate_file(expected_backup_folder,
                  '2023-07-07T15:02:10-tar.tar')  # Expected to be kept, because retention will be set to 7
    generate_file(expected_backup_folder,
                  '2023-07-06T15:02:10-tar.tar')  # Expected to be kept, because retention will be set to 7
    # When
    return_code = main(['backup', 'file', '--config-file', str(config_file)])

    # Then
    assert_that(return_code).is_equal_to(0)
    output = []
    with os.scandir(expected_backup_folder) as it:
        entry: os.DirEntry
        for entry in it:
            output.append({'file-name': entry.name})

    assert_that(output).contains_only({'file-name': '2023-07-09T15:02:10-tar-diff-clean.tar'},
                                      {'file-name': '2023-07-09T15:02:10-tar-snap-w27.snar'},
                                      {'file-name': '2023-07-08T15:02:10-tar.tar'},
                                      {'file-name': '2023-07-07T15:02:11-tar.tar'},
                                      {'file-name': '2023-07-07T15:02:10-tar.tar'},
                                      {'file-name': '2023-07-06T15:02:10-tar.tar'})

    assert_that(caplog.record_tuples).contains(
        ('root', logging.WARNING, 'WARNING: retention [2] is lower than the minimum possible [7]. '
                                  'Retention value is override by the minimum'))


def generate_file(expected_backup_folder, file_name):
    if not os.path.isdir(expected_backup_folder):
        os.makedirs(expected_backup_folder)
    with open(expected_backup_folder / file_name, 'w') as f:
        f.write('content')



from typing import Dict, List

from bashckup.actuators.actuators import ActuatorMetadata
from bashckup.actuators.post_backup import CleanFolderPostBackup, RsyncPostBackup, AbstractPostBackup
from bashckup.actuators.readers import FileReader, MariaDBReader, AbstractReader
from bashckup.actuators.transformers import GzipTransformer, OpenSSLTransformer, AbstractTransformer
from bashckup.actuators.writers import FileWriter, AbstractWriter


class ActuatorFactory:
    readerModules = [FileReader, MariaDBReader]
    transformerModules = [GzipTransformer, OpenSSLTransformer]
    writerModules = [FileWriter]
    postBackupModules = [CleanFolderPostBackup, RsyncPostBackup]

    _reader = dict((x.module_name(), x.__name__) for x in readerModules)
    _transformers = dict((x.module_name(), x.__name__) for x in transformerModules)
    _writer = dict((x.module_name(), x.__name__) for x in writerModules)
    _post_backup = dict((x.module_name(), x.__name__) for x in postBackupModules)

    @staticmethod
    def reader_module_name() -> List[str]:
        return [x.module_name() for x in ActuatorFactory.readerModules]

    @staticmethod
    def transformer_module_name() -> List[str]:
        return [x.module_name() for x in ActuatorFactory.transformerModules]

    @staticmethod
    def writer_module_name() -> List[str]:
        return [x.module_name() for x in ActuatorFactory.writerModules]

    @staticmethod
    def post_backup_module_name() -> List[str]:
        return [x.module_name() for x in ActuatorFactory.postBackupModules]

    def build_reader(self, global_context: dict, config: dict,
                     metadata: Dict[str, Dict[str, ActuatorMetadata]]) -> AbstractReader:
        module_name = next(iter(config))
        module_class_name = self._reader.get(module_name)
        args = config[module_name]['args']
        if module_class_name is None:
            raise ValueError(f'''Module {module_name} is not managed''')
        else:
            return globals()[module_class_name](global_context, args, metadata)

    def build_transformer(self, global_context: dict, config: dict or str,
                          metadata: Dict[str, Dict[str, ActuatorMetadata]]) -> AbstractTransformer:
        if type(config) is dict:
            module_name = next(iter(config))
            args = config[module_name]['args']
        elif type(config) is str:
            module_name = config
            args = {}
        else:
            raise Exception(f'Type {type(config).__name__} is not handled')
        module_class_name = self._transformers.get(module_name)
        if module_class_name is None:
            raise ValueError(f'Module {module_name} is not managed')
        else:
            return globals()[module_class_name](global_context, args, metadata)

    def build_writer(self, global_context: dict, config: dict,
                     metadata: Dict[str, Dict[str, ActuatorMetadata]]) -> AbstractWriter:
        module_name = next(iter(config))
        module_class_name = self._writer.get(module_name)
        args = config[module_name]['args']
        if module_class_name is None:
            raise ValueError(f'''Module {module_name} is not managed''')
        else:
            return globals()[module_class_name](global_context, args, metadata)

    def build_post_backup(self, global_context: dict, config: dict or str,
                          metadata: Dict[str, Dict[str, ActuatorMetadata]]) -> AbstractPostBackup:
        if type(config) is dict:
            module_name = next(iter(config))
            args = config[module_name]['args']
        elif type(config) is str:
            module_name = config
            args = {}
        else:
            raise Exception(f'Type {type(config).__name__} is not handled')
        module_class_name = self._post_backup.get(module_name)
        if module_class_name is None:
            raise ValueError(f'Module {module_name} is not managed')
        else:
            return globals()[module_class_name](global_context, args, metadata)


from typing import List, Optional


DEFAULT_IGNORE_LIST = [
    'nose.plugins',
    'six.moves',
    'django.utils.six.moves',
    'google.gax',
    'threading',
    'Queue',
    'selenium',
    '_pytest.terminal.',
    '_pytest.runner.',
    'gi',
]


class Settings:
    def __init__(self, default_ignore_list: Optional[List[str]]=None) -> None:
        self.default_ignore_list = default_ignore_list or DEFAULT_IGNORE_LIST[:]


settings = Settings()


class ConfigurationError(Exception):
    pass


def configure(default_ignore_list: Optional[List[str]]=None, extend_ignore_list: Optional[List[str]]=None) -> None:
    if default_ignore_list is not None and extend_ignore_list is not None:
        raise ConfigurationError("Either default_ignore_list or extend_ignore_list might be given, not both")
    if default_ignore_list:
        settings.default_ignore_list = default_ignore_list
    if extend_ignore_list:
        settings.default_ignore_list = [*settings.default_ignore_list, *extend_ignore_list]


def reset_config() -> None:
    global settings
    settings = Settings()



# Copyright (c) 2015-2019, Activision Publishing, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import collections

if sys.version_info[0] == 3:
    Iterable = collections.abc.Iterable
else:
    Iterable = collections.Iterable

__tracebackhide__ = True


class DynamicMixin(object):
    """Dynamic assertions mixin.

    When testing attributes of an object (or the contents of a dict), the
    :meth:`~assertpy.base.BaseMixin.is_equal_to` assertion can be a bit verbose::

        fred = Person('Fred', 'Smith')

        assert_that(fred.first_name).is_equal_to('Fred')
        assert_that(fred.name).is_equal_to('Fred Smith')
        assert_that(fred.say_hello()).is_equal_to('Hello, Fred!')

    Instead, use dynamic assertions in the form of ``has_<name>()`` where ``<name>`` is the name of
    any attribute, property, or zero-argument method on the given object. Dynamic equality
    assertions test if actual is equal to expected using the ``==`` operator. Using dynamic
    assertions, we can rewrite the above example as::

        assert_that(fred).has_first_name('Fred')
        assert_that(fred).has_name('Fred Smith')
        assert_that(fred).has_say_hello('Hello, Fred!')

    Similarly, dynamic assertions also work on any *dict-like* object::

        fred = {
            'first_name': 'Fred',
            'last_name': 'Smith',
            'shoe_size': 12
        }

        assert_that(fred).has_first_name('Fred')
        assert_that(fred).has_last_name('Smith')
        assert_that(fred).has_shoe_size(12)
    """

    def __getattr__(self, attr):
        """Asserts that val has attribute attr and that its value is equal to other via a dynamic
        assertion of the form ``has_<attr>()``."""
        if not attr.startswith('has_'):
            raise AttributeError('assertpy has no assertion <%s()>' % attr)

        attr_name = attr[4:]
        err_msg = False
        is_dict = isinstance(self.val, Iterable) and hasattr(self.val, '__getitem__')

        if not hasattr(self.val, attr_name):
            if is_dict:
                if attr_name not in self.val:
                    err_msg = 'Expected key <%s>, but val has no key <%s>.' % (attr_name, attr_name)
            else:
                err_msg = 'Expected attribute <%s>, but val has no attribute <%s>.' % (attr_name, attr_name)

        def _wrapper(*args, **kwargs):
            if err_msg:
                self.error(err_msg)  # ok to raise AssertionError now that we are inside wrapper
            else:
                if len(args) != 1:
                    raise TypeError('assertion <%s()> takes exactly 1 argument (%d given)' % (attr, len(args)))

                if is_dict:
                    val_attr = self.val[attr_name]
                else:
                    val_attr = getattr(self.val, attr_name)

                if callable(val_attr):
                    try:
                        actual = val_attr()
                    except TypeError:
                        raise TypeError('val does not have zero-arg method <%s()>' % attr_name)
                else:
                    actual = val_attr

                expected = args[0]
                if actual != expected:
                    self.error('Expected <%s> to be equal to <%s> on %s <%s>, but was not.' % (actual, expected, 'key' if is_dict else 'attribute', attr_name))
            return self

        return _wrapper



import os
import subprocess
from abc import ABC
from pathlib import Path
from typing import IO, AnyStr

from jsonschema.validators import validate

from bashckup.actuators.actuators import CommandActuator
from bashckup.actuators.exceptions import ParameterException


class AbstractTransformer(CommandActuator, ABC):

    def _actuator_type(self) -> str:
        return 'transformer'

    def generate_backup_process(self, stdin: IO[AnyStr], stdout: IO[AnyStr] = subprocess.PIPE) \
            -> subprocess.Popen:
        return super().generate_backup_process(stdin, stdout)


class GzipTransformer(AbstractTransformer):
    defaultLevel = 6
    validation_schema = {'type': 'object',
                         'properties': {
                             'level': {
                                 'type': 'integer',
                                 'minimum': 1,
                                 'maximum': 9,
                                 'default': defaultLevel,
                                 'description': 'Regulate the speed of compression using the specified digit #, '
                                                'where 1 indicates the fastest compression method (less compression) '
                                                'and 9 indicates the slowest compression method (best compression). '
                                                'The default compression level is ' + str(defaultLevel) +
                                                ' (that is, biased towards high compression at expense of speed).'}
                         },
                         'additionalProperties': False}

    @staticmethod
    def module_name() -> str:
        return 'gzip'

    def _get_params(self) -> None:
        validate(self._args, self.validation_schema)

        self.level = self._args.get('level', self.defaultLevel)

    def _generate_backup_cmd(self) -> [str]:
        cmd = ['gzip', '-' + str(self.level)]
        return cmd

    def _generate_restore_cmd(self) -> [str]:
        cmd = ['unzip']
        return cmd


class OpenSSLTransformer(AbstractTransformer):
    defaultLevel = 6
    validation_schema = {'type': 'object',
                         'properties': {
                             'password-file': {
                                 'type': 'string',
                                 'description': 'Path to the file that contains the password'}
                         },
                         'required': ['password-file'],
                         'additionalProperties': False}

    @staticmethod
    def module_name() -> str:
        return 'crypt'

    def _get_params(self) -> None:
        validate(self._args, self.validation_schema)

        password_file = Path(self._args['password-file'])
        if not password_file.is_file():
            raise ParameterException(f'''File [{self._args['password-file']}] doesn't exists''',
                                     'password-file', self._backup_id, self.module_name())
        if not os.access(password_file, os.R_OK):
            raise ParameterException(f'''File [{self._args['password-file']}] is not readable''', 'password-file',
                                     self._backup_id, self.module_name())
        password_file_stat = os.stat(password_file)
        if password_file_stat.st_mode & 0o077 != 0o0:
            raise ParameterException(f'''File [{self._args['password-file']}] must not be readable from group and '''
                                     'from others', 'password-file',
                                     self._backup_id, self.module_name())
        if password_file_stat.st_uid != os.getuid():
            raise ParameterException(
                'File [' + self._args['password-file'] + '] is not owned by current user', 'password-file',
                self._backup_id, self.module_name())
        self.password_file = password_file

    def _generate_backup_cmd(self) -> [str]:
        cmd = ['openssl', 'enc', '-e', '-aes-256-cbc', '-pbkdf2', '-kfile', str(self.password_file)]
        return cmd

    def _generate_restore_cmd(self) -> [str]:
        cmd = ['openssl', 'enc', '-d', '-aes-256-cbc', '-pbkdf2', '-kfile', str(self.password_file)]
        return cmd



# Copyright (c) 2015-2019, Activision Publishing, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import re
import collections

if sys.version_info[0] == 3:
    str_types = (str,)
    unicode = str
    Iterable = collections.abc.Iterable
else:
    str_types = (basestring,)
    unicode = unicode
    Iterable = collections.Iterable

__tracebackhide__ = True


class StringMixin(object):
    """String assertions mixin."""

    def is_equal_to_ignoring_case(self, other):
        """Asserts that val is a string and is case-insensitive equal to other.

        Checks actual is equal to expected using the ``==`` operator and ``str.lower()``.

        Args:
            other: the expected value

        Examples:
            Usage::

                assert_that('foo').is_equal_to_ignoring_case('FOO')
                assert_that('FOO').is_equal_to_ignoring_case('foo')
                assert_that('fOo').is_equal_to_ignoring_case('FoO')

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if actual is **not** case-insensitive equal to expected
        """
        if not isinstance(self.val, str_types):
            raise TypeError('val is not a string')
        if not isinstance(other, str_types):
            raise TypeError('given arg must be a string')
        if self.val.lower() != other.lower():
            self.error('Expected <%s> to be case-insensitive equal to <%s>, but was not.' % (self.val, other))
        return self

    def contains_ignoring_case(self, *items):
        """Asserts that val is string and contains the given item or items.

        Walks val and checks for item or items using the ``==`` operator and ``str.lower()``.

        Args:
            *items: the item or items expected to be contained

        Examples:
            Usage::

                assert_that('foo').contains_ignoring_case('F', 'oO')
                assert_that(['a', 'B']).contains_ignoring_case('A', 'b')
                assert_that({'a': 1, 'B': 2}).contains_ignoring_case('A', 'b')
                assert_that({'a', 'B'}).contains_ignoring_case('A', 'b')

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val does **not** contain the case-insensitive item or items
        """
        if len(items) == 0:
            raise ValueError('one or more args must be given')
        if isinstance(self.val, str_types):
            if len(items) == 1:
                if not isinstance(items[0], str_types):
                    raise TypeError('given arg must be a string')
                if items[0].lower() not in self.val.lower():
                    self.error('Expected <%s> to case-insensitive contain item <%s>, but did not.' % (self.val, items[0]))
            else:
                missing = []
                for i in items:
                    if not isinstance(i, str_types):
                        raise TypeError('given args must all be strings')
                    if i.lower() not in self.val.lower():
                        missing.append(i)
                if missing:
                    self.error('Expected <%s> to case-insensitive contain items %s, but did not contain %s.' % (
                        self.val, self._fmt_items(items), self._fmt_items(missing)))
        elif isinstance(self.val, Iterable):
            missing = []
            for i in items:
                if not isinstance(i, str_types):
                    raise TypeError('given args must all be strings')
                found = False
                for v in self.val:
                    if not isinstance(v, str_types):
                        raise TypeError('val items must all be strings')
                    if i.lower() == v.lower():
                        found = True
                        break
                if not found:
                    missing.append(i)
            if missing:
                self.error('Expected <%s> to case-insensitive contain items %s, but did not contain %s.' % (
                    self.val, self._fmt_items(items), self._fmt_items(missing)))
        else:
            raise TypeError('val is not a string or iterable')
        return self

    def starts_with(self, prefix):
        """Asserts that val is string or iterable and starts with prefix.

        Args:
            prefix: the prefix

        Examples:
            Usage::

                assert_that('foo').starts_with('fo')
                assert_that(['a', 'b', 'c']).starts_with('a')
                assert_that((1, 2, 3)).starts_with(1)
                assert_that(((1, 2), (3, 4), (5, 6))).starts_with((1, 2))

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val does **not** start with prefix
        """
        if prefix is None:
            raise TypeError('given prefix arg must not be none')
        if isinstance(self.val, str_types):
            if not isinstance(prefix, str_types):
                raise TypeError('given prefix arg must be a string')
            if len(prefix) == 0:
                raise ValueError('given prefix arg must not be empty')
            if not self.val.startswith(prefix):
                self.error('Expected <%s> to start with <%s>, but did not.' % (self.val, prefix))
        elif isinstance(self.val, Iterable):
            if len(self.val) == 0:
                raise ValueError('val must not be empty')
            first = next(iter(self.val))
            if first != prefix:
                self.error('Expected %s to start with <%s>, but did not.' % (self.val, prefix))
        else:
            raise TypeError('val is not a string or iterable')
        return self

    def ends_with(self, suffix):
        """Asserts that val is string or iterable and ends with suffix.

        Args:
            suffix: the suffix

        Examples:
            Usage::

                assert_that('foo').ends_with('oo')
                assert_that(['a', 'b', 'c']).ends_with('c')
                assert_that((1, 2, 3)).ends_with(3)
                assert_that(((1, 2), (3, 4), (5, 6))).ends_with((5, 6))

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val does **not** end with suffix
        """
        if suffix is None:
            raise TypeError('given suffix arg must not be none')
        if isinstance(self.val, str_types):
            if not isinstance(suffix, str_types):
                raise TypeError('given suffix arg must be a string')
            if len(suffix) == 0:
                raise ValueError('given suffix arg must not be empty')
            if not self.val.endswith(suffix):
                self.error('Expected <%s> to end with <%s>, but did not.' % (self.val, suffix))
        elif isinstance(self.val, Iterable):
            if len(self.val) == 0:
                raise ValueError('val must not be empty')
            last = None
            for last in self.val:
                pass
            if last != suffix:
                self.error('Expected %s to end with <%s>, but did not.' % (self.val, suffix))
        else:
            raise TypeError('val is not a string or iterable')
        return self

    def matches(self, pattern):
        """Asserts that val is string and matches the given regex pattern.

        Args:
            pattern (str): the regular expression pattern, as raw string (aka prefixed with ``r``)

        Examples:
            Usage::

                assert_that('foo').matches(r'\\w')
                assert_that('123-456-7890').matches(r'\\d{3}-\\d{3}-\\d{4}')

            Match is partial unless anchored, so these assertion pass::

                assert_that('foo').matches(r'\\w')
                assert_that('foo').matches(r'oo')
                assert_that('foo').matches(r'\\w{2}')

            To match the entire string, just use an anchored regex pattern where ``^`` and ``$``
            match the start and end of line and ``\\A`` and ``\\Z`` match the start and end of string::

                assert_that('foo').matches(r'^\\w{3}$')
                assert_that('foo').matches(r'\\A\\w{3}\\Z')

            And regex flags, such as ``re.MULTILINE`` and ``re.DOTALL``, can only be applied via
            *inline modifiers*, such as ``(?m)`` and ``(?s)``::

                s = '''bar
                foo
                baz'''

                # using multiline (?m)
                assert_that(s).matches(r'(?m)^foo$')

                # using dotall (?s)
                assert_that(s).matches(r'(?s)b(.*)z')

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val does **not** match pattern

        Tip:
            Regular expressions are tricky.  Be sure to use raw strings (aka prefixed with ``r``).
            Also, note that the :meth:`matches` assertion passes for partial matches (as does the
            underlying ``re.match`` method).  So, if you need to match the entire string, you must
            include anchors in the regex pattern.
        """
        if not isinstance(self.val, str_types):
            raise TypeError('val is not a string')
        if not isinstance(pattern, str_types):
            raise TypeError('given pattern arg must be a string')
        if len(pattern) == 0:
            raise ValueError('given pattern arg must not be empty')
        if re.search(pattern, self.val) is None:
            self.error('Expected <%s> to match pattern <%s>, but did not.' % (self.val, pattern))
        return self

    def does_not_match(self, pattern):
        """Asserts that val is string and does not match the given regex pattern.

        Args:
            pattern (str): the regular expression pattern, as raw string (aka prefixed with ``r``)

        Examples:
            Usage::

                assert_that('foo').does_not_match(r'\\d+')
                assert_that('123').does_not_match(r'\\w+')

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val **does** match pattern

        See Also:
            :meth:`matches` - for more about regex patterns
        """
        if not isinstance(self.val, str_types):
            raise TypeError('val is not a string')
        if not isinstance(pattern, str_types):
            raise TypeError('given pattern arg must be a string')
        if len(pattern) == 0:
            raise ValueError('given pattern arg must not be empty')
        if re.search(pattern, self.val) is not None:
            self.error('Expected <%s> to not match pattern <%s>, but did.' % (self.val, pattern))
        return self

    def is_alpha(self):
        """Asserts that val is non-empty string and all characters are alphabetic (using ``str.isalpha()``).

        Examples:
            Usage::

                assert_that('foo').is_lower()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** lowercase
        """
        if not isinstance(self.val, str_types):
            raise TypeError('val is not a string')
        if len(self.val) == 0:
            raise ValueError('val is empty')
        if not self.val.isalpha():
            self.error('Expected <%s> to contain only alphabetic chars, but did not.' % self.val)
        return self

    def is_digit(self):
        """Asserts that val is non-empty string and all characters are digits (using ``str.isdigit()``).

        Examples:
            Usage::

                assert_that('1234567890').is_digit()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** digits
        """
        if not isinstance(self.val, str_types):
            raise TypeError('val is not a string')
        if len(self.val) == 0:
            raise ValueError('val is empty')
        if not self.val.isdigit():
            self.error('Expected <%s> to contain only digits, but did not.' % self.val)
        return self

    def is_lower(self):
        """Asserts that val is non-empty string and all characters are lowercase (using ``str.lower()``).

        Examples:
            Usage::

                assert_that('foo').is_lower()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** lowercase
        """
        if not isinstance(self.val, str_types):
            raise TypeError('val is not a string')
        if len(self.val) == 0:
            raise ValueError('val is empty')
        if self.val != self.val.lower():
            self.error('Expected <%s> to contain only lowercase chars, but did not.' % self.val)
        return self

    def is_upper(self):
        """Asserts that val is non-empty string and all characters are uppercase (using ``str.upper()``).

        Examples:
            Usage::

                assert_that('FOO').is_upper()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** uppercase
        """
        if not isinstance(self.val, str_types):
            raise TypeError('val is not a string')
        if len(self.val) == 0:
            raise ValueError('val is empty')
        if self.val != self.val.upper():
            self.error('Expected <%s> to contain only uppercase chars, but did not.' % self.val)
        return self

    def is_unicode(self):
        """Asserts that val is a unicode string.

        Examples:
            Usage::

                assert_that(u'foo').is_unicode()  # python 2
                assert_that('foo').is_unicode()   # python 3

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** a unicode string
        """
        if type(self.val) is not unicode:
            self.error('Expected <%s> to be unicode, but was <%s>.' % (self.val, type(self.val).__name__))
        return self



# Copyright (c) 2015-2019, Activision Publishing, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import division
import sys
import math
import numbers
import datetime

__tracebackhide__ = True


class NumericMixin(object):
    """Numeric assertions mixin."""

    _NUMERIC_COMPAREABLE = set([datetime.datetime, datetime.timedelta, datetime.date, datetime.time])
    _NUMERIC_NON_COMPAREABLE = set([complex])

    def _validate_compareable(self, other):
        self_type = type(self.val)
        other_type = type(other)

        if self_type in self._NUMERIC_NON_COMPAREABLE:
            raise TypeError('ordering is not defined for type <%s>' % self_type.__name__)
        if self_type in self._NUMERIC_COMPAREABLE:
            if other_type is not self_type:
                raise TypeError('given arg must be <%s>, but was <%s>' % (self_type.__name__, other_type.__name__))
            return
        if isinstance(self.val, numbers.Number):
            if not isinstance(other, numbers.Number):
                raise TypeError('given arg must be a number, but was <%s>' % other_type.__name__)
            return
        raise TypeError('ordering is not defined for type <%s>' % self_type.__name__)

    def _validate_number(self):
        """Raise TypeError if val is not numeric."""
        if isinstance(self.val, numbers.Number) is False:
            raise TypeError('val is not numeric')

    def _validate_real(self):
        """Raise TypeError if val is not real number."""
        if isinstance(self.val, numbers.Real) is False:
            raise TypeError('val is not real number')

    def is_zero(self):
        """Asserts that val is numeric and is zero.

        Examples:
            Usage::

                assert_that(0).is_zero()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** zero
        """
        self._validate_number()
        return self.is_equal_to(0)

    def is_not_zero(self):
        """Asserts that val is numeric and is *not* zero.

        Examples:
            Usage::

                assert_that(1).is_not_zero()
                assert_that(123.4).is_not_zero()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val **is** zero
        """
        self._validate_number()
        return self.is_not_equal_to(0)

    def is_nan(self):
        """Asserts that val is real number and is ``NaN`` (not a number).

        Examples:
            Usage::

                assert_that(float('nan')).is_nan()
                assert_that(float('inf') * 0).is_nan()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** NaN
        """
        self._validate_number()
        self._validate_real()
        if not math.isnan(self.val):
            self.error('Expected <%s> to be <NaN>, but was not.' % self.val)
        return self

    def is_not_nan(self):
        """Asserts that val is real number and is *not* ``NaN`` (not a number).

        Examples:
            Usage::

                assert_that(0).is_not_nan()
                assert_that(123.4).is_not_nan()
                assert_that(float('inf')).is_not_nan()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val **is** NaN
        """
        self._validate_number()
        self._validate_real()
        if math.isnan(self.val):
            self.error('Expected not <NaN>, but was.')
        return self

    def is_inf(self):
        """Asserts that val is real number and is ``Inf`` (infinity).

        Examples:
            Usage::

                assert_that(float('inf')).is_inf()
                assert_that(float('inf') * 1).is_inf()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** Inf
        """
        self._validate_number()
        self._validate_real()
        if not math.isinf(self.val):
            self.error('Expected <%s> to be <Inf>, but was not.' % self.val)
        return self

    def is_not_inf(self):
        """Asserts that val is real number and is *not* ``Inf`` (infinity).

        Examples:
            Usage::

                assert_that(0).is_not_inf()
                assert_that(123.4).is_not_inf()
                assert_that(float('nan')).is_not_inf()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val **is** Inf
        """
        self._validate_number()
        self._validate_real()
        if math.isinf(self.val):
            self.error('Expected not <Inf>, but was.')
        return self

    def is_greater_than(self, other):
        """Asserts that val is numeric and is greater than other.

        Args:
            other: the other date, expected to be less than val

        Examples:
            Usage::

                assert_that(1).is_greater_than(0)
                assert_that(123.4).is_greater_than(111.1)

            For dates, behavior is identical to :meth:`~assertpy.date.DateMixin.is_after`::

                import datetime

                today = datetime.datetime.now()
                yesterday = today - datetime.timedelta(days=1)

                assert_that(today).is_greater_than(yesterday)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** greater than other
        """
        self._validate_compareable(other)
        if self.val <= other:
            if type(self.val) is datetime.datetime:
                self.error('Expected <%s> to be greater than <%s>, but was not.' % (
                    self.val.strftime('%Y-%m-%d %H:%M:%S'), other.strftime('%Y-%m-%d %H:%M:%S')))
            else:
                self.error('Expected <%s> to be greater than <%s>, but was not.' % (self.val, other))
        return self

    def is_greater_than_or_equal_to(self, other):
        """Asserts that val is numeric and is greater than or equal to other.

        Args:
            other: the other date, expected to be less than or equal to val

        Examples:
            Usage::

                assert_that(1).is_greater_than_or_equal_to(0)
                assert_that(1).is_greater_than_or_equal_to(1)
                assert_that(123.4).is_greater_than_or_equal_to(111.1)

            For dates, behavior is identical to :meth:`~assertpy.date.DateMixin.is_after` *except* when equal::

                import datetime

                today = datetime.datetime.now()
                yesterday = today - datetime.timedelta(days=1)

                assert_that(today).is_greater_than_or_equal_to(yesterday)
                assert_that(today).is_greater_than_or_equal_to(today)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** greater than or equal to other
        """
        self._validate_compareable(other)
        if self.val < other:
            if type(self.val) is datetime.datetime:
                self.error('Expected <%s> to be greater than or equal to <%s>, but was not.' % (
                    self.val.strftime('%Y-%m-%d %H:%M:%S'), other.strftime('%Y-%m-%d %H:%M:%S')))
            else:
                self.error('Expected <%s> to be greater than or equal to <%s>, but was not.' % (self.val, other))
        return self

    def is_less_than(self, other):
        """Asserts that val is numeric and is less than other.

        Args:
            other: the other date, expected to be greater than val

        Examples:
            Usage::

                assert_that(0).is_less_than(1)
                assert_that(123.4).is_less_than(555.5)

            For dates, behavior is identical to :meth:`~assertpy.date.DateMixin.is_before`::

                import datetime

                today = datetime.datetime.now()
                yesterday = today - datetime.timedelta(days=1)

                assert_that(yesterday).is_less_than(today)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** less than other
        """
        self._validate_compareable(other)
        if self.val >= other:
            if type(self.val) is datetime.datetime:
                self.error('Expected <%s> to be less than <%s>, but was not.' % (self.val.strftime('%Y-%m-%d %H:%M:%S'), other.strftime('%Y-%m-%d %H:%M:%S')))
            else:
                self.error('Expected <%s> to be less than <%s>, but was not.' % (self.val, other))
        return self

    def is_less_than_or_equal_to(self, other):
        """Asserts that val is numeric and is less than or equal to other.

        Args:
            other: the other date, expected to be greater than or equal to val

        Examples:
            Usage::

                assert_that(1).is_less_than_or_equal_to(0)
                assert_that(1).is_less_than_or_equal_to(1)
                assert_that(123.4).is_less_than_or_equal_to(100.0)

            For dates, behavior is identical to :meth:`~assertpy.date.DateMixin.is_before` *except* when equal::

                import datetime

                today = datetime.datetime.now()
                yesterday = today - datetime.timedelta(days=1)

                assert_that(yesterday).is_less_than_or_equal_to(today)
                assert_that(today).is_less_than_or_equal_to(today)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** less than or equal to other
        """
        self._validate_compareable(other)
        if self.val > other:
            if type(self.val) is datetime.datetime:
                self.error('Expected <%s> to be less than or equal to <%s>, but was not.' % (
                    self.val.strftime('%Y-%m-%d %H:%M:%S'), other.strftime('%Y-%m-%d %H:%M:%S')))
            else:
                self.error('Expected <%s> to be less than or equal to <%s>, but was not.' % (self.val, other))
        return self

    def is_positive(self):
        """Asserts that val is numeric and is greater than zero.

        Examples:
            Usage::

                assert_that(1).is_positive()
                assert_that(123.4).is_positive()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** positive
        """
        return self.is_greater_than(0)

    def is_negative(self):
        """Asserts that val is numeric and is less than zero.

        Examples:
            Usage::

                assert_that(-1).is_negative()
                assert_that(-123.4).is_negative()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** negative
        """
        return self.is_less_than(0)

    def is_between(self, low, high):
        """Asserts that val is numeric and is between low and high.

        Args:
            low: the low value
            high: the high value

        Examples:
            Usage::

                assert_that(1).is_between(0, 2)
                assert_that(123.4).is_between(111.1, 222.2)

            For dates, works as expected::

                import datetime

                today = datetime.datetime.now()
                middle = today - datetime.timedelta(hours=12)
                yesterday = today - datetime.timedelta(days=1)

                assert_that(middle).is_between(yesterday, today)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** between low and high
        """
        val_type = type(self.val)
        self._validate_between_args(val_type, low, high)

        if self.val < low or self.val > high:
            if val_type is datetime.datetime:
                self.error('Expected <%s> to be between <%s> and <%s>, but was not.' % (
                    self.val.strftime('%Y-%m-%d %H:%M:%S'), low.strftime('%Y-%m-%d %H:%M:%S'), high.strftime('%Y-%m-%d %H:%M:%S')))
            else:
                self.error('Expected <%s> to be between <%s> and <%s>, but was not.' % (self.val, low, high))
        return self

    def is_not_between(self, low, high):
        """Asserts that val is numeric and is *not* between low and high.

        Args:
            low: the low value
            high: the high value

        Examples:
            Usage::

                assert_that(1).is_not_between(2, 3)
                assert_that(1.1).is_not_between(2.2, 3.3)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val **is** between low and high
        """
        val_type = type(self.val)
        self._validate_between_args(val_type, low, high)

        if self.val >= low and self.val <= high:
            if val_type is datetime.datetime:
                self.error('Expected <%s> to not be between <%s> and <%s>, but was.' % (
                    self.val.strftime('%Y-%m-%d %H:%M:%S'), low.strftime('%Y-%m-%d %H:%M:%S'), high.strftime('%Y-%m-%d %H:%M:%S')))
            else:
                self.error('Expected <%s> to not be between <%s> and <%s>, but was.' % (self.val, low, high))
        return self

    def is_close_to(self, other, tolerance):
        """Asserts that val is numeric and is close to other within tolerance.

        Args:
            other: the other value, expected to be close to val within tolerance
            tolerance: the tolerance

        Examples:
            Usage::

                assert_that(123).is_close_to(100, 25)
                assert_that(123.4).is_close_to(123, 0.5)

            For dates, works as expected::

                import datetime

                today = datetime.datetime.now()
                yesterday = today - datetime.timedelta(days=1)

                assert_that(today).is_close_to(yesterday, datetime.timedelta(hours=36))

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** close to other within tolerance
        """
        self._validate_close_to_args(self.val, other, tolerance)

        if self.val < (other-tolerance) or self.val > (other+tolerance):
            if type(self.val) is datetime.datetime:
                tolerance_seconds = tolerance.days * 86400 + tolerance.seconds + tolerance.microseconds / 1000000
                h, rem = divmod(tolerance_seconds, 3600)
                m, s = divmod(rem, 60)
                self.error('Expected <%s> to be close to <%s> within tolerance <%d:%02d:%02d>, but was not.' % (
                    self.val.strftime('%Y-%m-%d %H:%M:%S'), other.strftime('%Y-%m-%d %H:%M:%S'), h, m, s))
            else:
                self.error('Expected <%s> to be close to <%s> within tolerance <%s>, but was not.' % (self.val, other, tolerance))
        return self

    def is_not_close_to(self, other, tolerance):
        """Asserts that val is numeric and is *not* close to other within tolerance.

        Args:
            other: the other value
            tolerance: the tolerance

        Examples:
            Usage::

                assert_that(123).is_not_close_to(100, 22)
                assert_that(123.4).is_not_close_to(123, 0.1)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val **is** close to other within tolerance
        """
        self._validate_close_to_args(self.val, other, tolerance)

        if self.val >= (other-tolerance) and self.val <= (other+tolerance):
            if type(self.val) is datetime.datetime:
                tolerance_seconds = tolerance.days * 86400 + tolerance.seconds + tolerance.microseconds / 1000000
                h, rem = divmod(tolerance_seconds, 3600)
                m, s = divmod(rem, 60)
                self.error('Expected <%s> to not be close to <%s> within tolerance <%d:%02d:%02d>, but was.' % (
                    self.val.strftime('%Y-%m-%d %H:%M:%S'), other.strftime('%Y-%m-%d %H:%M:%S'), h, m, s))
            else:
                self.error('Expected <%s> to not be close to <%s> within tolerance <%s>, but was.' % (self.val, other, tolerance))
        return self



import locale
import os
import shutil
import subprocess
from pathlib import Path

from _pytest.fixtures import fixture
from assertpy import assert_that

current_path = Path(os.path.dirname(os.path.realpath(__file__)))
tests_path = current_path / '..' / '..'
conf_path = tests_path / 'resources' / 'confs'
locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')

"""
Run application from CLI to to do end to end test
"""


def test_cli_process(backup_folder, server_data_folder):
    """
    GOAL: Test application launch as a CLI process
    Based on tar test
    """
    # Given
    config_file = conf_path / 'tar.yml'
    # When
    process = subprocess.run(['bashckup', 'backup', 'file', '--config-file', str(config_file)], capture_output=True)

    # Then
    assert_that(process.returncode).is_equal_to(0)
    assert_that(process.stdout).is_not_empty()
    assert_that(process.stderr).is_empty()



import locale
import os
import shutil
from pathlib import Path

from _pytest.fixtures import fixture
from assertpy import assert_that
from freezegun import freeze_time

from bashckup.bashckup import main

current_path = Path(os.path.dirname(os.path.realpath(__file__)))
tests_path = current_path / '..' / '..'
conf_path = tests_path / 'resources' / 'confs'
locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')

"""
Depends on TAR and MARIADB tests
"""


@freeze_time('2023-07-10 15:02:10')
def test_mariadb_gz(backup_folder, server_data_folder):
    """
    GOAL: Test GZIP with default options (no args)
    """
    # Given
    config_file = conf_path / 'mariadb-gz.yml'
    expected_backup_folder = backup_folder / 'mariadb-gz'
    # When
    return_code = main(['backup', 'file', '--config-file', str(config_file)])

    # Then
    assert_that(return_code).is_equal_to(0)
    output = []
    with os.scandir(expected_backup_folder) as it:
        entry: os.DirEntry
        for entry in it:
            output.append({'file-name': entry.name, 'size': entry.stat().st_size})

    assert_that(output).is_length(1)
    assert_that(output[0]['file-name']).is_equal_to('2023-07-10T15:02:10-mariadb-gz.sql.gz')
    # Info: Without compression its 10240 octets
    assert_that(output[0]['size']).is_between(9000, 9200)


@freeze_time('2023-07-10 15:02:10')
def test_tar_gz(backup_folder, server_data_folder):
    """
    GOAL: Test GZIP with custom option (level set to 9)
    """
    # Given
    config_file = conf_path / 'tar-gz.yml'
    expected_backup_folder = backup_folder / 'tar-gz'
    # When
    return_code = main(['backup', 'file', '--config-file', str(config_file)])

    # Then
    assert_that(return_code).is_equal_to(0)
    output = []
    with os.scandir(expected_backup_folder) as it:
        entry: os.DirEntry
        for entry in it:
            output.append({'file-name': entry.name, 'size': entry.stat().st_size})

    assert_that(output).is_length(1)
    bck_file = output[0]
    assert_that(bck_file['file-name']).is_equal_to('2023-07-10T15:02:10-tar-gz.tar.gz')
    # With level set to 9 and same files, output size change a lot between executions
    # Info: Without compression its 10240 octets
    assert_that(bck_file['size']).is_between(160, 210)
